import contextlib
import dataclasses
import errno
import logging
import posixpath
import re
import struct
import time
from collections import namedtuple
from collections.abc import Generator
from datetime import datetime
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from construct import Array, Container, Int32ul
from parameter_decorators import path_to_str

from rpcclient.clients.darwin.consts import (
    EXC_MASK_ALL,
    EXC_TYPES_COUNT,
    MACH_PORT_TYPE_ALL_RIGHTS,
    MACH_PORT_TYPE_DEAD_NAME,
    MACH_PORT_TYPE_DNREQUEST,
    MACH_PORT_TYPE_PORT_SET,
    MACH_PORT_TYPE_RECEIVE,
    MACH_PORT_TYPE_SEND,
    MACH_PORT_TYPE_SEND_ONCE,
    TASK_DYLD_INFO,
    TASK_FLAVOR_READ,
    TASK_VM_INFO,
    THREAD_IDENTIFIER_INFO,
    VM_FLAGS_ANYWHERE,
    ARMThreadFlavors,
    kCSNow,
    x86_THREAD_STATE64,
)
from rpcclient.clients.darwin.structs import (
    ARM_THREAD_STATE64_COUNT,
    FAT_CIGAM,
    FAT_MAGIC,
    LOAD_COMMAND_TYPE,
    MAXPATHLEN,
    PROC_PIDFDPIPEINFO,
    PROC_PIDFDSOCKETINFO,
    PROC_PIDFDVNODEPATHINFO,
    PROC_PIDLISTFDS,
    PROC_PIDTASKALLINFO,
    PROX_FDTYPE_KQUEUE,
    PROX_FDTYPE_PIPE,
    PROX_FDTYPE_SOCKET,
    PROX_FDTYPE_VNODE,
    TASK_DYLD_INFO_COUNT,
    TASK_VM_INFO_COUNT,
    THREAD_IDENTIFIER_INFO_COUNT,
    all_image_infos_t,
    arm_thread_state64_t,
    dyld_image_info_t,
    fat_header,
    ipc_info_name_t,
    mach_header_t,
    mach_port_t,
    pid_t,
    pipe_info,
    proc_fdinfo,
    proc_taskallinfo,
    procargs2_t,
    so_family_t,
    so_kind_t,
    socket_fdinfo,
    task_dyld_info_data_t,
    task_vm_info_data_t,
    thread_identifier_info,
    vnode_fdinfowithpath,
    x86_thread_state64_t,
)
from rpcclient.clients.darwin.symbol import DarwinSymbol
from rpcclient.core.structs.consts import SEEK_SET, SIGKILL, SIGTERM
from rpcclient.core.subsystems.processes import Processes
from rpcclient.core.subsystems.sysctl import CTL, KERN
from rpcclient.core.symbol import ADDRESS_SIZE_TO_STRUCT_FORMAT, Symbol
from rpcclient.exceptions import (
    ArgumentError,
    BadReturnValueError,
    ProcessSymbolAbsentError,
    RpcClientException,
    SymbolAbsentError,
    UnrecognizedSelectorError,
)
from rpcclient.protos.rpc_pb2 import ARCH_ARM64

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient

_CF_STRING_ARRAY_PREFIX_LEN = len('    "')
_CF_STRING_ARRAY_SUFFIX_LEN = len('",')
_BACKTRACE_FRAME_REGEX = re.compile(r"\[\s*(\d+)\] (0x[0-9a-f]+)\s+\{(.+?) \+ (.+?)\} (.*)")

FdStruct = namedtuple("FdStruct", "fd struct")

logger = logging.getLogger(__name__)

CDHASH_SIZE = 20
CHUNK_SIZE = 1024 * 64
APP_SUFFIX = ".app/"


@dataclasses.dataclass()
class Fd:
    fd: int


@dataclasses.dataclass()
class KQueueFd(Fd):
    pass


@dataclasses.dataclass()
class PipeFd(Fd):
    pass


@dataclasses.dataclass()
class FileFd(Fd):
    path: str


@dataclasses.dataclass()
class UnixFd(Fd):
    path: str


@dataclasses.dataclass()
class SocketFd(Fd):
    pass


@dataclasses.dataclass()
class Ipv4SocketFd(SocketFd):
    local_address: str
    local_port: int
    remote_address: str
    remote_port: int  # when remote 0, the socket is for listening


@dataclasses.dataclass()
class Ipv6SocketFd(SocketFd):
    local_address: str
    local_port: int
    remote_address: str
    remote_port: int  # when remote 0, the socket is for listening


@dataclasses.dataclass()
class Ipv4TcpFd(Ipv4SocketFd):
    pass


@dataclasses.dataclass()
class Ipv6TcpFd(Ipv6SocketFd):
    pass


@dataclasses.dataclass()
class Ipv4UdpFd(Ipv4SocketFd):
    pass


@dataclasses.dataclass()
class Ipv6UdpFd(Ipv6SocketFd):
    pass


Image = namedtuple("Image", "address path")

SOCKET_TYPE_DATACLASS = {
    so_family_t.AF_INET: {
        so_kind_t.SOCKINFO_TCP: Ipv4TcpFd,
        so_kind_t.SOCKINFO_IN: Ipv4UdpFd,
    },
    so_family_t.AF_INET6: {
        so_kind_t.SOCKINFO_TCP: Ipv6TcpFd,
        so_kind_t.SOCKINFO_IN: Ipv6UdpFd,
    },
}


class Thread:
    def __init__(self, client: "DarwinClient", thread_id: int):
        """Initialize a thread wrapper for the given client and thread id."""
        self._client = client
        self._thread_id = thread_id

    @property
    def thread_id(self) -> int:
        """Return the Mach thread id."""
        return self._thread_id

    def get_state(self):
        """Return the architecture-specific thread state."""
        raise NotImplementedError()

    def set_state(self, state: dict):
        """Set the architecture-specific thread state."""
        raise NotImplementedError()

    def resume(self):
        """Resume execution of the thread."""
        raise NotImplementedError()

    def suspend(self):
        """Suspend execution of the thread."""
        raise NotImplementedError()

    def __repr__(self):
        """Return a debug representation of the thread."""
        return f"<{self.__class__.__name__} TID:{self._thread_id}>"


class IntelThread64(Thread):
    def get_state(self):
        """Fetch the x86_64 thread state."""
        with (
            self._client.safe_malloc(x86_thread_state64_t.sizeof()) as p_state,
            self._client.safe_malloc(x86_thread_state64_t.sizeof()) as p_thread_state_count,
        ):
            p_thread_state_count[0] = x86_thread_state64_t.sizeof() // Int32ul.sizeof()
            if self._client.symbols.thread_get_state(
                self._thread_id, x86_THREAD_STATE64, p_state, p_thread_state_count
            ):
                raise BadReturnValueError("thread_get_state() failed")
            return x86_thread_state64_t.parse_stream(p_state)

    def set_state(self, state: dict) -> None:
        """Set the x86_64 thread state."""
        if self._client.symbols.thread_set_state(
            self._thread_id,
            x86_THREAD_STATE64,
            x86_thread_state64_t.build(state),
            x86_thread_state64_t.sizeof() // Int32ul.sizeof(),
        ):
            raise BadReturnValueError("thread_set_state() failed")


class ArmThread64(Thread):
    def get_state(self):
        """Fetch the ARM64 thread state."""
        with (
            self._client.safe_malloc(arm_thread_state64_t.sizeof()) as p_state,
            self._client.safe_malloc(arm_thread_state64_t.sizeof()) as p_thread_state_count,
        ):
            p_thread_state_count[0] = ARM_THREAD_STATE64_COUNT
            if self._client.symbols.thread_get_state(
                self._thread_id, ARMThreadFlavors.ARM_THREAD_STATE64, p_state, p_thread_state_count
            ):
                raise BadReturnValueError("thread_get_state() failed")
            return arm_thread_state64_t.parse_stream(p_state)

    def set_state(self, state: dict) -> None:
        """Set the ARM64 thread state."""
        if self._client.symbols.thread_set_state(
            self._thread_id,
            ARMThreadFlavors.ARM_THREAD_STATE64,
            arm_thread_state64_t.build(state),
            ARM_THREAD_STATE64_COUNT,
        ):
            raise BadReturnValueError("thread_set_state() failed")

    def suspend(self):
        """Suspend execution of the ARM64 thread."""
        if self._client.symbols.thread_suspend(self._thread_id):
            raise BadReturnValueError("thread_suspend() failed")

    def resume(self):
        """Resume execution of the ARM64 thread."""
        if self._client.symbols.thread_resume(self._thread_id):
            raise BadReturnValueError("thread_resume() failed")


@dataclasses.dataclass
class Region:
    region_type: str
    start: "ProcessSymbol"
    end: int
    vsize: str
    protection: str
    protection_max: str
    region_detail: str

    @property
    def size(self) -> int:
        """Return the region size in bytes."""
        return self.end - self.start


@dataclasses.dataclass
class LoadedClass:
    name: str
    type_name: str
    binary_path: str


@dataclasses.dataclass
class Frame:
    depth: int
    address: int
    section: str
    offset: int
    symbol_name: str

    def __repr__(self):
        """Return a formatted frame representation."""
        return (
            f"<{self.__class__.__name__} [{self.depth:3}] 0x{self.address:x} ({self.section} + 0x{self.offset:x}) "
            f"{self.symbol_name}>"
        )


@dataclasses.dataclass
class Backtrace:
    flavor: str
    time_start: float
    time_end: float
    pid: int
    thread_id: int
    dispatch_queue_serial_num: int
    frames: list[Frame]

    def __init__(self, vmu_backtrace: DarwinSymbol):
        """Parse a VMU backtrace description into structured fields."""
        backtrace = vmu_backtrace.objc_call("description").py()
        match = re.match(
            r"VMUBacktrace \(Flavor: (?P<flavor>.+?) Simple Time: (?P<time>.+?) "
            r"Process: (?P<pid>\d+) Thread: (?P<thread_id>.+?)  Dispatch queue serial num: "
            r"(?P<dispatch_queue_serial_num>\d+)\)",
            backtrace,
        )
        self.flavor = match.group("flavor")
        self.pid = int(match.group("pid"))
        self.thread_id = int(match.group("thread_id"), 16)
        self.dispatch_queue_serial_num = int(match.group("dispatch_queue_serial_num"))

        self.frames = []
        for frame in re.findall(_BACKTRACE_FRAME_REGEX, backtrace):
            self.frames.append(
                Frame(
                    depth=int(frame[0]),
                    address=int(frame[1], 0),
                    section=frame[2],
                    offset=int(frame[3], 0),
                    symbol_name=frame[4],
                )
            )

    def __repr__(self):
        """Return a formatted backtrace representation."""
        buf = f"<{self.__class__.__name__} PID: {self.pid} TID: {self.thread_id}\n"
        for frame in self.frames:
            buf += f"    {frame}\n"
        buf += ">"
        return buf


class ProcessSymbol(Symbol):
    @classmethod
    def create(cls, value: int, client, process):
        """Create a ProcessSymbol bound to a process."""
        symbol = super().create(value, client)
        symbol = ProcessSymbol(symbol)
        symbol._prepare(client)
        symbol.process = process
        return symbol

    def _symbol_from_value(self, value: int):
        """Clone this symbol for the given value."""
        return self.create(value, self._client, self.process)

    def peek(self, count: int) -> bytes:
        """Read bytes from process memory at this address."""
        return self.process.peek(self, count)

    def poke(self, buf: bytes) -> None:
        """Write bytes to process memory at this address."""
        return self.process.poke(self, buf)

    def peek_str(self, encoding="utf-8") -> str:
        """peek string at given address"""
        return self.process.peek_str(self, encoding)

    @property
    def dl_info(self) -> Container:
        """Raise because dl_info is not available for remote process symbols."""
        raise NotImplementedError("dl_info isn't implemented for remote process symbols")

    @property
    def name(self) -> str:
        """Return the symbol name for this address."""
        return self.process.get_symbol_name(self)

    @property
    def filename(self) -> str:
        """Return the image path containing this symbol."""
        return self.process.get_symbol_image(self.name).path

    def __getitem__(self, item) -> "ProcessSymbol":
        """Dereference and return the pointer at the given index."""
        fmt = ADDRESS_SIZE_TO_STRUCT_FORMAT[self.item_size]
        addr = self + item * self.item_size
        deref = struct.unpack(self._client._endianness + fmt, self.process.peek(addr, self.item_size))[0]
        return self.create(deref, self._client, self.process)

    def __setitem__(self, item, value):
        """Write a pointer-sized value at the given index."""
        fmt = ADDRESS_SIZE_TO_STRUCT_FORMAT[self.item_size]
        value = struct.pack(self._client._endianness + fmt, int(value))
        self.process.poke(self + item * self.item_size, value)

    def __call__(self, *args, **kwargs):
        """Disallow calling a process symbol as a function."""
        raise RpcClientException("ProcessSymbol is not callable")


@dataclasses.dataclass
class MachPortThreadInfo:
    thread_ids: list[int]


@dataclasses.dataclass
class MachPortInfo:
    task: int
    pid: int
    name: int
    rights: list[str]
    ipc_object: int
    dead: bool
    proc_name: Optional[str] = None
    thread_info: Optional[MachPortThreadInfo] = None

    @property
    def has_recv_right(self) -> bool:
        """Return True if the port has a receive right."""
        return "recv" in self.rights

    @property
    def has_send_right(self) -> bool:
        """Return True if the port has a send right."""
        return "send" in self.rights


@dataclasses.dataclass
class MachPortCrossRefInfo:
    name: int
    ipc_object: int
    recv_right_pid: int
    recv_right_proc_name: str


class SymbolOwner:
    def __init__(
        self, client: "DarwinClient", process, symbol_owner_opaque1: ProcessSymbol, symbol_owner_opaque2: ProcessSymbol
    ) -> None:
        """Initialize a symbol owner wrapper."""
        self._client = client
        self._process = process
        self.symbol_owner_opaque1 = symbol_owner_opaque1
        self.symbol_owner_opaque2 = symbol_owner_opaque2

    def get_symbol_address(self, name: str) -> ProcessSymbol:
        """Resolve a symbol address by name."""
        symbol = self._client.symbols.CSSymbolOwnerGetSymbolWithName(
            self.symbol_owner_opaque1, self.symbol_owner_opaque2, name, return_raw=True
        )
        return ProcessSymbol.create(
            self._client.symbols.CSSymbolGetRange(symbol.x0, symbol.x1), self._client, self._process
        )


class Symbolicator:
    def __init__(
        self, client: "DarwinClient", process, symbolicator_opaque1: ProcessSymbol, symbolicator_opaque2: ProcessSymbol
    ) -> None:
        """Initialize a symbolicator wrapper."""
        self._client = client
        self._process = process
        self.symbolicator_opaque1 = symbolicator_opaque1
        self.symbolicator_opaque2 = symbolicator_opaque2

    def get_symbol_owner(self, library_basename: str) -> SymbolOwner:
        """Resolve a symbol owner by library basename."""
        symbol_owner = self._client.symbols.CSSymbolicatorGetSymbolOwnerWithNameAtTime(
            self.symbolicator_opaque1, self.symbolicator_opaque2, library_basename, kCSNow, return_raw=True
        )
        return SymbolOwner(self._client, self._process, symbol_owner.x0, symbol_owner.x1)


class Process:
    PEEK_STR_CHUNK_SIZE = 0x100

    def __init__(self, client: "DarwinClient", pid: int):
        """Initialize a process wrapper for a pid."""
        self._client = client
        self._pid = pid

        if self._client.arch == ARCH_ARM64:
            self._thread_class = ArmThread64
        else:
            self._thread_class = IntelThread64

    def kill(self, sig: int = SIGTERM):
        """Send a signal to the remote process."""
        return self._client.processes.kill(self._pid, sig)

    def waitpid(self, flags: int = 0):
        """Wait for the remote process to change state and return status."""
        return self._client.processes.waitpid(self._pid, flags)

    def peek(self, address: int, size: int) -> bytes:
        """peek at memory address"""
        with self._client.safe_malloc(size) as buf, self._client.safe_malloc(8) as p_size:
            p_size[0] = size
            if self._client.symbols.vm_read_overwrite(self.task_read, address, size, buf, p_size):
                raise BadReturnValueError("vm_read() failed")
            return buf.peek(size)

    def peek_str(self, address: int, encoding="utf-8") -> str:
        """peek string at memory address"""
        size = self.PEEK_STR_CHUNK_SIZE
        buf = b""

        while size:
            try:
                buf += self.peek(address, size)
                if b"\x00" in buf:
                    return buf.split(b"\x00", 1)[0].decode(encoding)
                address += size
            except BadReturnValueError:
                size = size // 2

    def poke(self, address: int, buf: bytes):
        """poke at memory address"""
        if self._client.symbols.vm_write(self.task, address, buf, len(buf)):
            raise BadReturnValueError("vm_write() failed")

    def get_symbol_name(self, address: int) -> str:
        """Resolve a symbol name for the given address."""
        if self._client.arch != ARCH_ARM64:
            raise NotImplementedError("implemented only on ARCH_ARM64")
        result = self.vmu_object_identifier.objc_call("symbolForAddress:", address, return_raw=True)
        if result.x0 == 0 and result.x1 == 0:
            raise SymbolAbsentError()
        return self._client.symbols.CSSymbolGetName(result.x0, result.x1).peek_str()

    def get_symbol_image(self, name: str) -> Image:
        """Find the image containing the named symbol."""
        for image in self.images:
            result = self.get_symbol_address(name, posixpath.basename(image.path))
            if result:
                return image

        raise ProcessSymbolAbsentError()

    def get_symbol_class_info(self, address: int) -> DarwinSymbol:
        """Return Objective-C class info for the given address."""
        return self.vmu_object_identifier.objc_call("classInfoForMemory:length:", address, 8)

    def get_symbol_address(self, name: str, lib: Optional[str] = None) -> Union[DarwinSymbol, ProcessSymbol]:
        """Resolve a symbol address, optionally within a library."""
        if lib is not None:
            address = self.vmu_object_identifier.objc_call("addressOfSymbol:inLibrary:", name, lib).c_uint64
            if self.pid == self._client.pid:
                return self._client.symbol(address)
            return ProcessSymbol.create(address, self._client, self)

        image = self.get_symbol_image(name)
        return self.get_symbol_address(name, posixpath.basename(image.path))

    @property
    def loaded_classes(self):
        """Yield realized Objective-C classes for the process."""
        realized_classes = self.vmu_object_identifier.objc_call("realizedClasses")

        for i in range(1, realized_classes.objc_call("count") + 1):
            class_info = realized_classes.objc_call("classInfoForIndex:", i)
            name = class_info.objc_call("className").py()
            type_name = class_info.objc_call("typeName").py()
            binary_path = class_info.objc_call("binaryPath").py()
            yield LoadedClass(name=name, type_name=type_name, binary_path=binary_path)

    @property
    def images(self) -> list[Image]:
        """get loaded image list"""
        result = []

        with (
            self._client.safe_malloc(task_dyld_info_data_t.sizeof()) as dyld_info,
            self._client.safe_calloc(8) as count,
        ):
            count[0] = TASK_DYLD_INFO_COUNT
            if self._client.symbols.task_info(self.task_read, TASK_DYLD_INFO, dyld_info, count):
                raise BadReturnValueError("task_info(TASK_DYLD_INFO) failed")
            dyld_info_data = task_dyld_info_data_t.parse_stream(dyld_info)
        all_image_infos = all_image_infos_t.parse(
            self.peek(dyld_info_data.all_image_info_addr, dyld_info_data.all_image_info_size)
        )

        buf = self.peek(all_image_infos.infoArray, all_image_infos.infoArrayCount * dyld_image_info_t.sizeof())
        for image in Array(all_image_infos.infoArrayCount, dyld_image_info_t).parse(buf):
            path = self.peek_str(image.imageFilePath)
            result.append(Image(address=self.get_process_symbol(image.imageLoadAddress), path=path))
        return result

    @property
    def app_images(self) -> list[Image]:
        """Return images that belong to the app bundle."""
        return [image for image in self.images if APP_SUFFIX in image.path]

    @property
    def threads(self) -> list[Thread]:
        """Return the list of threads in the process."""
        result = []
        with self._client.safe_malloc(8) as threads, self._client.safe_malloc(4) as count:
            count.item_size = 4
            if self._client.symbols.task_threads(self.task_read, threads, count):
                raise BadReturnValueError("task_threads() failed")

            for tid in Array(count[0].c_uint32, Int32ul).parse(threads[0].peek(count[0] * 4)):
                result.append(self._thread_class(self._client, tid))
        return result

    @property
    def pid(self) -> int:
        """get pid"""
        return self._pid

    @property
    def fds(self) -> list[Fd]:
        """get a list of process opened file descriptors"""
        result = []
        for fdstruct in self.fd_structs:
            fd = fdstruct.fd
            parsed = fdstruct.struct

            if fd.proc_fdtype == PROX_FDTYPE_VNODE:
                result.append(FileFd(fd=fd.proc_fd, path=parsed.pvip.vip_path))

            elif fd.proc_fdtype == PROX_FDTYPE_KQUEUE:
                result.append(KQueueFd(fd=fd.proc_fd))

            elif fd.proc_fdtype == PROX_FDTYPE_PIPE:
                result.append(PipeFd(fd=fd.proc_fd))

            elif fd.proc_fdtype == PROX_FDTYPE_SOCKET:
                if parsed.psi.soi_kind in (so_kind_t.SOCKINFO_TCP, so_kind_t.SOCKINFO_IN):
                    correct_class = SOCKET_TYPE_DATACLASS[parsed.psi.soi_family][parsed.psi.soi_kind]

                    if parsed.psi.soi_kind == so_kind_t.SOCKINFO_TCP:
                        info = parsed.psi.soi_proto.pri_tcp.tcpsi_ini
                    else:
                        info = parsed.psi.soi_proto.pri_in
                    result.append(
                        correct_class(
                            fd=fd.proc_fd,
                            local_address=info.insi_laddr.ina_46.i46a_addr4,
                            local_port=info.insi_lport,
                            remote_address=info.insi_faddr.ina_46.i46a_addr4,
                            remote_port=info.insi_fport,
                        )
                    )

                elif parsed.psi.soi_kind == so_kind_t.SOCKINFO_UN:
                    result.append(UnixFd(fd=fd.proc_fd, path=parsed.psi.soi_proto.pri_un.unsi_addr.ua_sun.sun_path))

        return result

    @property
    def fd_structs(self) -> list[FdStruct]:
        """get a list of process opened file descriptors as raw structs"""
        result = []
        size = self._client.symbols.proc_pidinfo(self.pid, PROC_PIDLISTFDS, 0, 0, 0)

        vi_size = 8196  # should be enough for all structs
        with self._client.safe_malloc(vi_size) as vi_buf:
            with self._client.safe_malloc(size) as fdinfo_buf:
                size = int(self._client.symbols.proc_pidinfo(self.pid, PROC_PIDLISTFDS, 0, fdinfo_buf, size))
                if not size:
                    raise BadReturnValueError("proc_pidinfo(PROC_PIDLISTFDS) failed")

                for fd in Array(size // proc_fdinfo.sizeof(), proc_fdinfo).parse(fdinfo_buf.peek(size)):
                    if fd.proc_fdtype == PROX_FDTYPE_VNODE:
                        # file
                        vs = self._client.symbols.proc_pidfdinfo(
                            self.pid, fd.proc_fd, PROC_PIDFDVNODEPATHINFO, vi_buf, vi_size
                        )
                        if not vs:
                            if self._client.errno == errno.EBADF:
                                # lsof treats this as fine
                                continue
                            raise BadReturnValueError(
                                f"proc_pidinfo(PROC_PIDFDVNODEPATHINFO) failed for fd: {fd.proc_fd} "
                                f"({self._client.last_error})"
                            )

                        result.append(
                            FdStruct(
                                fd=fd, struct=vnode_fdinfowithpath.parse(vi_buf.peek(vnode_fdinfowithpath.sizeof()))
                            )
                        )

                    elif fd.proc_fdtype == PROX_FDTYPE_KQUEUE:
                        result.append(FdStruct(fd=fd, struct=None))

                    elif fd.proc_fdtype == PROX_FDTYPE_SOCKET:
                        # socket
                        vs = self._client.symbols.proc_pidfdinfo(
                            self.pid, fd.proc_fd, PROC_PIDFDSOCKETINFO, vi_buf, vi_size
                        )
                        if not vs:
                            if self._client.errno == errno.EBADF:
                                # lsof treats this as fine
                                continue
                            raise BadReturnValueError(
                                f"proc_pidinfo(PROC_PIDFDSOCKETINFO) failed ({self._client.last_error})"
                            )

                        result.append(FdStruct(fd=fd, struct=socket_fdinfo.parse(vi_buf.peek(vi_size))))

                    elif fd.proc_fdtype == PROX_FDTYPE_PIPE:
                        # pipe
                        vs = self._client.symbols.proc_pidfdinfo(
                            self.pid, fd.proc_fd, PROC_PIDFDPIPEINFO, vi_buf, vi_size
                        )
                        if not vs:
                            if self._client.errno == errno.EBADF:
                                # lsof treats this as fine
                                continue
                            raise BadReturnValueError(
                                f"proc_pidinfo(PROC_PIDFDPIPEINFO) failed ({self._client.last_error})"
                            )

                        result.append(FdStruct(fd=fd, struct=pipe_info.parse(vi_buf.peek(pipe_info.sizeof()))))

            return result

    @property
    def task_all_info(self):
        """get a list of process opened file descriptors"""
        with self._client.safe_malloc(proc_taskallinfo.sizeof()) as pti:
            if not self._client.symbols.proc_pidinfo(self.pid, PROC_PIDTASKALLINFO, 0, pti, proc_taskallinfo.sizeof()):
                raise BadReturnValueError("proc_pidinfo(PROC_PIDTASKALLINFO) failed")
            return proc_taskallinfo.parse_stream(pti)

    @property
    def task_vm_info(self) -> Container:
        """get TASK_VM_INFO via task_info."""
        with self._client.safe_malloc(task_vm_info_data_t.sizeof()) as vm_info, self._client.safe_calloc(8) as count:
            count[0] = TASK_VM_INFO_COUNT
            if self._client.symbols.task_info(self.task_read, TASK_VM_INFO, vm_info, count):
                raise BadReturnValueError("task_info(TASK_VM_INFO) failed")
            return task_vm_info_data_t.parse_stream(vm_info)

    @property
    def backtraces(self) -> list[Backtrace]:
        """Collect backtraces for all threads in the process."""
        result = []
        backtraces = self._client.symbols.objc_getClass("VMUSampler").objc_call(
            "sampleAllThreadsOfTask:", self.task_read
        )
        for i in range(backtraces.objc_call("count")):
            bt = backtraces.objc_call("objectAtIndex:", i)
            result.append(Backtrace(bt))
        return result

    @cached_property
    def vmu_proc_info(self) -> DarwinSymbol:
        """Return the VMUProcInfo object for this task."""
        return (
            self._client.symbols.objc_getClass("VMUProcInfo")
            .objc_call("alloc")
            .objc_call("initWithTask:", self.task_read)
        )

    @cached_property
    def vmu_region_identifier(self) -> DarwinSymbol:
        """Return the VMUVMRegionIdentifier object for this task."""
        return (
            self._client.symbols.objc_getClass("VMUVMRegionIdentifier")
            .objc_call("alloc")
            .objc_call("initWithTask:", self.task_read)
        )

    @cached_property
    def vmu_object_identifier(self) -> DarwinSymbol:
        """Return the VMUObjectIdentifier object for this task."""
        return (
            self._client.symbols.objc_getClass("VMUObjectIdentifier")
            .objc_call("alloc")
            .objc_call("initWithTask:", self.task_read)
        )

    @cached_property
    def symbolicator(self) -> Symbolicator:
        """Return a symbolicator for this task."""
        symbolicator = self._client.symbols.CSSymbolicatorCreateWithTask(self.task_read, return_raw=True)
        return Symbolicator(self._client, self, symbolicator.x0, symbolicator.x1)

    @cached_property
    def task_read(self) -> int:
        """Return the task read port for this process."""
        self_task_port = self._client.symbols.mach_task_self()

        if self.pid == self._client.pid:
            return self_task_port

        with self._client.safe_malloc(8) as p_task:
            ret = self._client.symbols.task_read_for_pid(self_task_port, self.pid, p_task)
            if ret:
                raise BadReturnValueError("task_read_for_pid() failed")
            return p_task[0].c_int64

    @cached_property
    def task(self) -> int:
        """Return the full task port for this process."""
        self_task_port = self._client.symbols.mach_task_self()

        if self.pid == self._client.pid:
            return self_task_port

        with self._client.safe_malloc(8) as p_task:
            if self._client.symbols.task_for_pid(self_task_port, self.pid, p_task):
                raise BadReturnValueError("task_for_pid() failed")
            return p_task[0].c_int64

    @cached_property
    def path(self) -> Optional[str]:
        """call proc_pidpath(filename, ...) at remote. review xnu header for more details."""
        with self._client.safe_malloc(MAXPATHLEN) as path:
            path_len = self._client.symbols.proc_pidpath(self.pid, path, MAXPATHLEN)
            if not path_len:
                return None
            return path.peek(path_len).decode()

    @cached_property
    def basename(self) -> Optional[str]:
        """Return the basename of the process path."""
        path = self.path
        if not path:
            return None
        return Path(path).parts[-1]

    @cached_property
    def name(self) -> str:
        """Return the process name from task info."""
        return self.task_all_info.pbsd.pbi_name

    @cached_property
    def ppid(self) -> int:
        """Return the parent process id."""
        return self.task_all_info.pbsd.pbi_ppid

    @cached_property
    def uid(self) -> int:
        """Return the process user id."""
        return self.task_all_info.pbsd.pbi_uid

    @cached_property
    def gid(self) -> int:
        """Return the process group id."""
        return self.task_all_info.pbsd.pbi_gid

    @cached_property
    def ruid(self) -> int:
        """Return the real user id."""
        return self.task_all_info.pbsd.pbi_ruid

    @cached_property
    def rgid(self) -> int:
        """Return the real group id."""
        return self.task_all_info.pbsd.pbi_rgid

    @cached_property
    def start_time(self) -> datetime:
        """Return the process start time."""
        if self._client.arch != ARCH_ARM64:
            raise NotImplementedError("implemented only on ARCH_ARM64")
        val = self.vmu_proc_info.objc_call("startTime", return_raw=True)
        tv_sec = val.x0
        tv_nsec = val.x1
        return datetime.fromtimestamp(tv_sec + (tv_nsec / (10**9)))

    @property
    def parent(self) -> "Process":
        """Return a Process wrapper for the parent pid."""
        return Process(self._client, self.ppid)

    @property
    def environ(self) -> list[str]:
        """Return the process environment variables."""
        return self.vmu_proc_info.objc_call("envVars").py()

    @property
    def arguments(self) -> list[str]:
        """Return the process argument list."""
        return self.vmu_proc_info.objc_call("arguments").py()

    @property
    def raw_procargs2(self) -> bytes:
        """Return the raw PROCARGS2 sysctl buffer."""
        return self._client.sysctl.get(CTL.KERN, KERN.PROCARGS2, self.pid)

    @property
    def procargs2(self) -> Container:
        """Parse and return the PROCARGS2 buffer."""
        return procargs2_t.parse(self.raw_procargs2)

    @property
    def regions(self) -> list[Region]:
        """Return the list of VM regions for the process."""
        result = []

        # remove the '()' wrapping the list:
        #   (
        #       "item1",
        #       "item2",
        #   )
        buf = self.vmu_region_identifier.objc_call("regions").cfdesc.split("\n")[1:-1]

        for line in buf:
            # remove line prefix and suffix and split into words
            line = line[_CF_STRING_ARRAY_PREFIX_LEN:-_CF_STRING_ARRAY_SUFFIX_LEN].split()
            i = 0

            region_type = line[i]
            i += 1

            while "-" not in line[i]:
                # skip till address range
                i += 1

            address_range = line[i]
            i += 1
            if not address_range.startswith("0x"):
                continue
            start, end = address_range.split("-")

            start = int(start, 16)
            end = int(end, 16)
            vsize = None
            if "V=" in line[i]:
                vsize = line[i].split("V=", 1)[1].split("]", 1)[0]

            while "/" not in line[i]:
                i += 1

            protection = line[i].split("/")
            i += 1

            region_detail = None
            if len(line) >= i + 1:
                region_detail = line[i]

            result.append(
                Region(
                    region_type=region_type,
                    start=self.get_process_symbol(start),
                    end=end,
                    vsize=vsize,
                    protection=protection[0],
                    protection_max=protection[1],
                    region_detail=region_detail,
                )
            )

        return result

    @property
    def cdhash(self) -> bytes:
        """Return the code directory hash for the process."""
        with self._client.safe_malloc(CDHASH_SIZE) as cdhash:
            # by reversing online-auth-agent
            if self._client.symbols.csops(self.pid, 5, cdhash, CDHASH_SIZE) != 0:
                raise BadReturnValueError(f"failed to get cdhash for {self.pid}")
            return cdhash.peek(CDHASH_SIZE)

    def get_process_symbol(self, address: int) -> ProcessSymbol:
        """Create a process symbol for the given address."""
        return ProcessSymbol.create(address, self._client, self)

    def vm_allocate(self, size: int) -> ProcessSymbol:
        """Allocate memory in the target task and return its address."""
        with self._client.safe_malloc(8) as out_address:
            if self._client.symbols.vm_allocate(self.task, out_address, size, VM_FLAGS_ANYWHERE):
                raise BadReturnValueError("vm_allocate() failed")
            return self.get_process_symbol(out_address[0])

    @path_to_str("output_dir")
    def dump_app(self, output_dir: str, chunk_size=CHUNK_SIZE) -> None:
        """
        Based on:
        https://github.com/AloneMonkey/frida-ios-dump/blob/master/dump.js
        """
        for image in self.app_images:
            relative_name = image.path.split(APP_SUFFIX, 1)[1]
            output_file = Path(output_dir).expanduser() / relative_name
            output_file.parent.mkdir(exist_ok=True, parents=True)
            logger.debug(f"dumping: {output_file}")

            with open(output_file, "wb") as output_file:
                macho_in_memory = mach_header_t.parse_stream(image.address)

                with self._client.fs.open(image.path, "r") as fat_file:
                    # locating the correct MachO offset within the FAT image (if FAT)
                    magic_in_fs = Int32ul.parse_stream(fat_file)
                    fat_file.seek(0, SEEK_SET)
                    if magic_in_fs in (FAT_CIGAM, FAT_MAGIC):
                        parsed_fat = fat_header.parse_stream(fat_file)
                        for arch in parsed_fat.archs:
                            if (
                                arch.cputype == macho_in_memory.cputype
                                and arch.cpusubtype == macho_in_memory.cpusubtype
                            ):
                                # correct MachO offset found
                                file_offset = arch.offset
                                break
                        fat_file.seek(file_offset, SEEK_SET)

                    # perform actual MachO dump
                    output_file.seek(0, SEEK_SET)
                    while True:
                        chunk = fat_file.read(chunk_size)
                        if len(chunk) == 0:
                            break
                        output_file.write(chunk)

                    offset_cryptid = None

                    # if image is encrypted, patch its encryption loader command
                    for load_command in macho_in_memory.load_commands:
                        if load_command.cmd in (
                            LOAD_COMMAND_TYPE.LC_ENCRYPTION_INFO_64,
                            LOAD_COMMAND_TYPE.LC_ENCRYPTION_INFO,
                        ):
                            offset_cryptid = load_command.data.cryptid_offset - image.address
                            crypt_offset = load_command.data.cryptoff
                            crypt_size = load_command.data.cryptsize
                            break

                    if offset_cryptid is not None:
                        output_file.seek(offset_cryptid, SEEK_SET)
                        output_file.write(b"\x00" * 4)  # cryptid = 0
                        output_file.seek(crypt_offset, SEEK_SET)
                        output_file.flush()
                        output_file.write((image.address + crypt_offset).peek(crypt_size))

    def get_mach_port_cross_ref_info(self) -> list[MachPortCrossRefInfo]:
        """Get all allocated mach ports and cross-refs to get the recv right owner"""
        result = []
        own_ports = []
        cross_refs_info = {}
        for port in self._client.processes.get_mach_ports():
            if port.pid == self.pid:
                own_ports.append(port)
            if port.ipc_object not in cross_refs_info:
                cross_refs_info[port.ipc_object] = []
            cross_refs_info[port.ipc_object].append(port)

        for port in own_ports:
            for cross_ref_port_info in cross_refs_info[port.ipc_object]:
                if cross_ref_port_info.has_recv_right:
                    result.append(
                        MachPortCrossRefInfo(
                            name=port.name,
                            ipc_object=port.ipc_object,
                            recv_right_pid=cross_ref_port_info.pid,
                            recv_right_proc_name=cross_ref_port_info.proc_name,
                        )
                    )
        return result

    def get_memgraph_snapshot(self) -> bytes:
        """Return a memory graph snapshot as plist bytes."""
        scanner = None
        try:
            # first attempt via new API
            scanner = self._client.symbols.objc_getClass("VMUProcessObjectGraph").objc_call(
                "createWithTask:", self.task_read
            )
            snapshot_graph = scanner.objc_call("plistRepresentationWithOptions:", 0).py()
        except UnrecognizedSelectorError:
            # if failed, attempt with old API
            scanner = (
                self._client.symbols.objc_getClass("VMUTaskMemoryScanner")
                .objc_call("alloc")
                .objc_call("initWithTask:", self.task_read)
            )
            scanner.objc_call("addRootNodesFromTask")
            scanner.objc_call("addMallocNodesFromTask")
            snapshot_graph = (
                scanner.objc_call("processSnapshotGraph").objc_call("plistRepresentationWithOptions:", 0).py()
            )
        finally:
            if scanner is not None:
                # free object scanner so process can resume
                scanner.objc_call("release")
        return snapshot_graph

    def __repr__(self):
        """Return a debug representation of the process."""
        return f"<{self.__class__.__name__} PID:{self.pid} PATH:{self.path}>"


class DarwinProcesses(Processes):
    """manage processes"""

    def __init__(self, client: "DarwinClient"):
        """Initialize the Darwin process subsystem."""
        super().__init__(client)
        client.load_framework("Symbolication")
        self._self_process = Process(self._client, self._client.pid)

    def get_self(self) -> Process:
        """get self process"""
        return self._self_process

    def get_by_pid(self, pid: int) -> Process:
        """get process object by pid"""
        proc_list = self.list()
        for p in proc_list:
            if p.pid == pid:
                return p
        raise ArgumentError(f"failed to locate process with pid: {pid}")

    def get_by_basename(self, name: str) -> Process:
        """get process object by basename"""
        proc_list = self.list()
        for p in proc_list:
            if p.basename == name:
                return p
        raise ArgumentError(f"failed to locate process with name: {name}")

    def get_by_name(self, name: str) -> Process:
        """get process object by name"""
        proc_list = self.list()
        for p in proc_list:
            if p.name == name:
                return p
        raise ArgumentError(f"failed to locate process with name: {name}")

    def grep(self, name: str) -> list[Process]:
        """get process list by basename filter"""
        result = []
        proc_list = self.list()
        for p in proc_list:
            if p.basename and name in p.basename:
                result.append(p)
        return result

    def get_processes_by_listening_port(self, port: int) -> list[Process]:
        """get a process object listening on the specified port"""
        listening_processes = []
        for process in self.list():
            try:
                fds = process.fds
            except BadReturnValueError:
                # it's possible to get error if new processes have since died or the rpcserver
                # doesn't have the required permissions to access all the processes
                continue

            for fd in fds:
                if (isinstance(fd, (Ipv4SocketFd, Ipv6SocketFd))) and fd.local_port == port and fd.remote_port == 0:
                    listening_processes.append(process)
        return listening_processes

    def lsof(self) -> dict[int, list[Fd]]:
        """get dictionary of pid to its opened fds"""
        result = {}
        for process in self.list():
            try:
                fds = process.fds
            except BadReturnValueError:
                # it's possible to get error if new processes have since died or the rpcserver
                # doesn't have the required permissions to access all the processes
                continue

            result[process.pid] = fds
        return result

    @path_to_str("path")
    def fuser(self, path: str) -> list[Process]:
        """get a list of all processes have an open hande to the specified path"""
        result = []
        proc_list = self.list()
        for process in proc_list:
            try:
                fds = process.fds
            except BadReturnValueError:
                # it's possible to get error if new processes have since died or the rpcserver
                # doesn't have the required permissions to access all the processes
                continue

            for fd in fds:
                if isinstance(fd, FileFd) and (str(Path(fd.path).absolute()) == str(Path(path).absolute())):
                    result.append(process)

        return result

    def list(self) -> list[Process]:
        """list all currently running processes"""
        n = self._client.symbols.proc_listallpids(0, 0)
        pid_buf_size = pid_t.sizeof() * n
        with self._client.safe_malloc(pid_buf_size) as pid_buf:
            pid_buf.item_size = pid_t.sizeof()
            n = self._client.symbols.proc_listallpids(pid_buf, pid_buf_size)

            result = []
            for i in range(n):
                pid = int(pid_buf[i])
                result.append(Process(self._client, pid))
            return result

    def disable_watchdog(self) -> None:
        """Continuously kill watchdogd to keep it disabled."""
        while True:
            with contextlib.suppress(ArgumentError):
                self.get_by_basename("watchdogd").kill(SIGKILL)
            time.sleep(1)

    def get_mach_ports(self, include_thread_info: bool = False) -> Generator[MachPortInfo, None, None]:
        """Enumerate all mach ports (heavily inspired by lsmp)"""
        thread_info = None
        mach_task_self = self._client.symbols.mach_task_self()

        p_psets = self._client.symbols.malloc(8)
        p_pset_count = self._client.symbols.malloc(8)
        p_pset_priv = self._client.symbols.malloc(8)
        p_tasks = self._client.symbols.malloc(8)
        p_task_count = self._client.symbols.malloc(8)
        p_count = self._client.symbols.malloc(4)
        ports_info = self._client.symbols.malloc(4 * 2 * EXC_TYPES_COUNT)
        masks = self._client.symbols.malloc(4 * EXC_TYPES_COUNT)
        behaviors = self._client.symbols.malloc(4 * EXC_TYPES_COUNT)
        flavors = self._client.symbols.malloc(4 * EXC_TYPES_COUNT)
        p_thread_count = self._client.symbols.malloc(4)
        p_thread_ports = self._client.symbols.malloc(8)
        info = self._client.symbols.malloc(200)
        p_pid = self._client.symbols.malloc(4)
        th_info = self._client.symbols.malloc(thread_identifier_info.sizeof())
        p_th_kobject = self._client.symbols.malloc(4)
        p_th_kotype = self._client.symbols.malloc(4)
        p_th_info_count = self._client.symbols.malloc(4)
        p_th_voucher = self._client.symbols.calloc(4, 1)
        p_table = self._client.symbols.malloc(8)
        unused = self._client.symbols.malloc(200)
        proc_name = self._client.symbols.malloc(100)
        p_kotype = self._client.symbols.malloc(4)

        p_pid.item_size = 4
        p_count.item_size = 4
        p_thread_count.item_size = 4
        p_th_info_count.item_size = 4
        p_th_voucher.item_size = 4
        p_th_info_count[0] = THREAD_IDENTIFIER_INFO_COUNT
        p_kotype.item_size = 4

        if self._client.symbols.getuid() == 0:
            # if privileged, get the info for all tasks so we can match ports up
            if (
                self._client.symbols.host_processor_sets(self._client.symbols.mach_host_self(), p_psets, p_pset_count)
                != 0
            ):
                raise BadReturnValueError("host_processor_sets() failed")
            if p_pset_count[0] != 1:
                raise BadReturnValueError("Assertion Failure: pset count greater than one")

            # convert the processor-set-name port to a privileged port
            if (
                self._client.symbols.host_processor_set_priv(
                    self._client.symbols.mach_host_self(), p_psets[0][0], p_pset_priv
                )
                != 0
            ):
                raise BadReturnValueError("host_processor_set_priv() failed")

            self._client.symbols.mach_port_deallocate(mach_task_self, p_psets[0][0])
            self._client.symbols.vm_deallocate(mach_task_self, p_psets[0], p_pset_count[0] * mach_port_t.sizeof())

            # convert the processor-set-priv to a list of task read ports for the processor set
            if (
                self._client.symbols.processor_set_tasks_with_flavor(
                    p_pset_priv[0], TASK_FLAVOR_READ, p_tasks, p_task_count
                )
                != 0
            ):
                raise BadReturnValueError("processor_set_tasks_with_flavor() failed")

            self._client.symbols.vm_deallocate(mach_task_self, p_pset_priv[0])

            # swap my current instances port to be last to collect all threads and exception port info
            my_task_position = None
            task_count = p_task_count[0]
            tasks = list(struct.unpack(f"<{int(task_count)}I", p_tasks[0].peek(task_count * 4)))

            for i in range(task_count):
                if self._client.symbols.mach_task_is_self(tasks[i]):
                    my_task_position = i
                    break
            if my_task_position is not None:
                swap_holder = tasks[task_count - 1]
                tasks[task_count - 1] = tasks[my_task_position]
                tasks[my_task_position] = swap_holder
        else:
            logger.warning("should run as root for best output (cross-ref to other tasks' ports)")
            # just the one process
            task_count = 1
            with self._client.safe_malloc(8) as p_task:
                ret = self._client.symbols.task_read_for_pid(mach_task_self, self.pid, p_task)
                if ret != 0:
                    raise BadReturnValueError("task_read_for_pid() failed")
                tasks = [p_task[0]]

        for task in tasks:
            self._client.symbols.pid_for_task(task, p_pid)
            pid = p_pid[0].c_uint32

            if (
                self._client.symbols.task_get_exception_ports_info(
                    task, EXC_MASK_ALL, masks, p_count, ports_info, behaviors, flavors
                )
                != 0
            ):
                raise BadReturnValueError("task_get_exception_ports_info() failed")

            if include_thread_info:
                # collect threads port as well
                if self._client.symbols.task_threads(task, p_thread_ports, p_thread_count) != 0:
                    raise BadReturnValueError("task_threads() failed")

                # collect the thread information
                thread_count = p_thread_count[0].c_uint32
                thread_ports = struct.unpack(f"<{thread_count}I", p_thread_ports[0].peek(4 * thread_count))
                thread_ids = []
                for thread_port in thread_ports:
                    ret = self._client.symbols.thread_get_exception_ports_info(
                        thread_port, EXC_MASK_ALL, masks, p_count, ports_info, behaviors, flavors
                    )
                    if ret != 0:
                        raise BadReturnValueError(
                            f"thread_get_exception_ports_info() failed: "
                            f"{self._client.symbols.mach_error_string(ret).peek_str()}"
                        )

                    if (
                        self._client.symbols.mach_port_kernel_object(
                            mach_task_self, thread_port, p_th_kotype, p_th_kobject
                        )
                        == 0
                    ) and (
                        self._client.symbols.thread_info(
                            mach_task_self, thread_port, THREAD_IDENTIFIER_INFO, th_info, p_th_info_count[0]
                        )
                        == 0
                    ):
                        thread_id = thread_identifier_info.parse(
                            th_info.peek(thread_identifier_info.sizeof())
                        ).thread_id
                        thread_ids.append(thread_id)

                    self._client.symbols.mach_port_deallocate(mach_task_self, thread_port)
                thread_info = MachPortThreadInfo(thread_ids=thread_ids)

            if self._client.symbols.mach_port_space_info(task, info, p_table, p_count, unused, unused) != 0:
                raise BadReturnValueError("mach_port_space_info() failed")

            proc_name_str = proc_name.peek_str() if self._client.symbols.proc_name(pid, proc_name, 100) != 0 else None

            count = p_count[0].c_uint32
            table_struct = Array(count, ipc_info_name_t)

            parsed_table = table_struct.parse(p_table[0].peek(table_struct.sizeof()))

            for entry in parsed_table:
                dnreq = False
                rights = []

                if entry.iin_type & MACH_PORT_TYPE_ALL_RIGHTS == 0:
                    # skip empty slots in the table
                    continue

                if entry.iin_type == MACH_PORT_TYPE_PORT_SET:
                    continue

                if entry.iin_type & MACH_PORT_TYPE_SEND:
                    rights.append("send")

                if entry.iin_type & MACH_PORT_TYPE_DNREQUEST:
                    dnreq = True

                if entry.iin_type & MACH_PORT_TYPE_RECEIVE:
                    rights.append("recv")

                elif entry.iin_type == MACH_PORT_TYPE_DEAD_NAME:
                    continue

                if entry.iin_type == MACH_PORT_TYPE_SEND_ONCE:
                    pass

                yield MachPortInfo(
                    task=task,
                    pid=pid,
                    name=entry.iin_name,
                    rights=rights,
                    ipc_object=entry.iin_object,
                    dead=dnreq,
                    proc_name=proc_name_str,
                    thread_info=thread_info,
                )



import contextlib
import logging
import os
import posixpath
import stat
import tempfile
from collections.abc import Generator
from pathlib import Path, PosixPath
from random import Random
from typing import TYPE_CHECKING, Union

from parameter_decorators import path_to_str

from rpcclient.clients.darwin.structs import MAXPATHLEN
from rpcclient.clients.darwin.symbol import DarwinSymbol
from rpcclient.core.allocated import Allocated
from rpcclient.core.structs.consts import (
    DT_DIR,
    DT_LNK,
    DT_REG,
    DT_UNKNOWN,
    O_CREAT,
    O_RDONLY,
    O_RDWR,
    O_TRUNC,
    O_WRONLY,
    R_OK,
    S_IFDIR,
    S_IFLNK,
    S_IFMT,
    S_IFREG,
    SEEK_CUR,
)
from rpcclient.exceptions import (
    ArgumentError,
    BadReturnValueError,
    RpcFileExistsError,
    RpcFileNotFoundError,
    RpcIsADirectoryError,
)

if TYPE_CHECKING:
    from rpcclient.core.client import CoreClient

logger = logging.getLogger(__name__)


class DirEntry:
    def __init__(self, path, entry, client: "CoreClient"):
        self._path = path
        self._entry = entry
        self._client = client
        self._lstat = None
        self._stat = None

    @property
    def name(self) -> str:
        return self._entry.d_name

    @property
    def path(self) -> str:
        return posixpath.join(self._path, self.name)

    def inode(self):
        """Return inode of the entry; cached per entry."""
        return self._entry.d_ino

    def is_dir(self, *, follow_symlinks=True):
        """Return True if the entry is a directory; cached per entry."""
        return self._test_mode(follow_symlinks, S_IFDIR)

    def is_file(self, *, follow_symlinks=True):
        """Return True if the entry is a file; cached per entry."""
        return self._test_mode(follow_symlinks, S_IFREG)

    def is_symlink(self):
        """Return True if the entry is a symbolic link; cached per entry."""
        if self._entry.d_type != DT_UNKNOWN:
            return self._entry.d_type == DT_LNK
        else:
            return self._test_mode(False, S_IFLNK)

    def stat(self, *, follow_symlinks=True):
        """Return stat_result object for the entry; cached per entry."""
        if not follow_symlinks:
            return self._get_lstat()
        if self._stat is None:
            if self.is_symlink():
                self._stat = self._fetch_stat(True)
            else:
                self._stat = self._get_lstat()
        return self._stat

    def _test_mode(self, follow_symlinks, mode):
        is_symlink = self._entry.d_type == DT_LNK
        need_stat = self._entry.d_type == DT_UNKNOWN or (follow_symlinks and is_symlink)
        if not need_stat:
            d_types = {S_IFLNK: DT_LNK, S_IFDIR: DT_DIR, S_IFREG: DT_REG}
            return self._entry.d_type == d_types[mode]
        else:
            st_mode = self.stat(follow_symlinks=follow_symlinks).st_mode
            if not st_mode:
                self._client.raise_errno_exception(f"failed to stat(): {self.path}")
            return (st_mode & S_IFMT) == mode

    def _get_lstat(self):
        if self._lstat is None:
            self._lstat = self._fetch_stat(False)
        return self._lstat

    def _fetch_stat(self, follow_symlinks):
        result = self._entry.stat
        if not follow_symlinks:
            result = self._entry.lstat

        if result.errno != 0:
            self._client.errno = result.errno
            self._client.raise_errno_exception(f"failed to stat: {self._entry.d_name}")
        return result

    def __repr__(self):
        return f"<{self.__class__.__name__} NAME:{self.name}{'/' if self.is_dir() else ''}>"


class File(Allocated):
    CHUNK_SIZE = 1024 * 64

    def __init__(self, client: "CoreClient", fd: int):
        """
        :param rpcclient.client.client.Client client:
        :param fd:
        """
        super().__init__()
        self._client = client
        self.fd = fd

    def _deallocate(self):
        """Close the remote file descriptor."""
        fd = self._client.symbols.close(self.fd).c_int32
        if fd < 0:
            self._client.raise_errno_exception(f"failed to close fd: {fd}")

    def seek(self, offset: int, whence: int) -> int:
        """Seek the remote file descriptor and return the resulting offset."""
        err = self._client.symbols.lseek(self.fd, offset, whence).c_int32
        if err < 0:
            self._client.raise_errno_exception(f"failed to lseek fd: {self.fd}")
        return err

    def tell(self) -> int:
        return self.seek(0, SEEK_CUR)

    def _write(self, buf: bytes) -> int:
        """Write bytes to the remote file descriptor."""
        n = self._client.symbols.write(self.fd, buf, len(buf)).c_int64
        if n < 0:
            self._client.raise_errno_exception(f"failed to write on fd: {self.fd}")
        return n

    def write(self, buf: bytes) -> int:
        """Write the full buffer, retrying until all bytes are written."""
        while buf:
            err = self._write(buf)
            buf = buf[err:]
        return len(buf)

    def _read(self, buf: DarwinSymbol, size: int) -> bytes:
        """read file at remote"""
        err = self._client.symbols.read(self.fd, buf, size).c_int64
        if err < 0:
            self._client.raise_errno_exception(f"read() failed for fd: {self.fd}")
        return buf.peek(err)

    def read_using_chunk(self, chunk: DarwinSymbol, chunk_size: int, size: int) -> bytes:
        buf = b""
        while size == -1 or len(buf) < size:
            read_chunk = self._read(chunk, chunk_size)
            if not read_chunk:
                # EOF
                break
            buf += read_chunk
        return buf

    def read(self, size: int = -1, chunk_size: int = CHUNK_SIZE, chunk: DarwinSymbol = None) -> bytes:
        """read file at remote"""
        if size != -1 and size < chunk_size:
            chunk_size = size

        buf = b""
        if chunk:
            return self.read_using_chunk(chunk, chunk_size, size)
        else:
            with self._client.safe_malloc(chunk_size) as temp_chunk:
                return self.read_using_chunk(temp_chunk, chunk_size, size)
        return buf

    def pread(self, length: int, offset: int) -> bytes:
        """call pread() at remote"""
        with self._client.safe_malloc(length) as buf:
            err = self._client.symbols.pread(self.fd, buf, length, offset).c_int64
            if err < 0:
                self._client.raise_errno_exception(f"pread() failed for fd: {self.fd}")
            return buf.peek(err)

    def pwrite(self, buf: bytes, offset: int):
        """call pwrite() at remote"""
        err = self._client.symbols.pwrite(self.fd, buf, len(buf), offset).c_int64
        if err < 0:
            self._client.raise_errno_exception(f"pwrite() failed for fd: {self.fd}")

    def fdatasync(self):
        err = self._client.symbols.fdatasync(self.fd).c_int64
        if err < 0:
            self._client.raise_errno_exception(f"fdatasync() failed for fd: {self.fd}")

    def fsync(self):
        err = self._client.symbols.fsync(self.fd).c_int64
        if err < 0:
            self._client.raise_errno_exception(f"fsync() failed for fd: {self.fd}")

    def flock(self, operation: int) -> None:
        """Apply or clear an advisory lock on the remote file descriptor."""
        err = self._client.symbols.flock(self.fd, operation).c_int32
        if err < 0:
            self._client.raise_errno_exception(f"flock() failed for fd: {self.fd}")

    def dup(self) -> int:
        err = self._client.symbols.dup(self.fd).c_int64
        if err < 0:
            self._client.raise_errno_exception(f"dup() failed for fd: {self.fd}")
        return err

    def __repr__(self):
        return f"<{self.__class__.__name__} FD:{self.fd}>"


class RemotePath(PosixPath):
    def __init__(self, path: str, client: "CoreClient") -> None:
        try:
            super().__init__(path)  # solution from python3.12 since signature has changed
        except TypeError:
            super().__init__()
        self._path = path
        self._client = client

    def __new__(cls, path: str, client):
        # this will not be needed once python3.11 is deprecated since it is now possible to subclass normally
        return super().__new__(cls, *[path])

    def chmod(self, mode: int):
        return self._client.fs.chmod(self._path, mode)

    def readlink(self):
        return self._client.fs.readlink(self._path)

    def exists(self) -> bool:
        try:
            self.stat()
        except Exception:
            return False
        return True

    def is_dir(self) -> bool:
        return bool(self.stat().st_mode & S_IFDIR)

    def lstat(self):
        return self._client.fs.lstat(self._path)

    def mkdir(self, mode: int, exist_ok=False):
        self._client.fs.mkdir(self._path, mode, exist_ok)

    def read_bytes(self) -> bytes:
        with self._open("r") as f:
            return f.read()

    def stat(self):
        return self._client.fs.stat(self._path)

    def symlink_to(self, target: Path, target_is_directory: bool = False) -> None:
        return self._client.fs.symlink(target, self._path)

    def write_bytes(self, buf: bytes) -> int:
        with self._open("w") as f:
            return f.write(buf)

    def _open(self, mode: str, access: int = 0o777) -> File:
        return self._client.fs.open(self._path, mode, access)

    def iterdir(self) -> Generator["RemotePath", None, None]:
        for entry in self._client.fs.listdir(self._path):
            yield self.__class__(f"{self._path}/{entry}", self._client)

    def touch(self, mode: int = 438, exist_ok: bool = True) -> None:
        try:
            return self._client.fs.touch(self._path, mode, exist_ok)
        except RpcFileExistsError as e:
            raise FileExistsError() from e

    def __truediv__(self, key: Path) -> "RemotePath":
        return RemotePath(f"{self._path}/{key}", self._client)

    def remove(self, recursive: bool = False, force: bool = False):
        """Remove the current path from the filesystem.

        :param recursive:
        :param force:
        """
        if self.exists():
            self._client.fs.remove(self._path, recursive=recursive, force=force)


class RemoteTemporaryDir(Allocated, RemotePath):
    """Temporary directory on the remote device"""

    def __init__(self, client: "CoreClient", directory: str = "/tmp", mode: int = 0o700):
        """Generate a random temp directory name and create it in the given directory (default '/tmp').

        :param rpcclient.darwin.client.DarwinClient client:
        :param directory: Directory to create the temp directory inside (default '/tmp').
        :param mode: Mode to create the temp directory with (default 0o700).
        """
        Allocated.__init__(self)
        remote_dir = client.fs.remote_path(directory)
        if not remote_dir.exists:
            raise RpcFileNotFoundError(f"remote dir {remote_dir} does not exist")
        name = "tmpdir_" + "".join(Random().choices("abcdefghijklmnopqrstuvwxyz0123456789_", k=8))
        remote_path = client.fs.remote_path(str(remote_dir / Path(name)))
        RemotePath.__init__(self, remote_path, client)
        self.mkdir(mode)

    def __new__(cls, *args, **kwargs):
        # this will not be needed once python3.11 is deprecated since it is now possible to subclass normally
        return RemotePath.__new__(cls, *[args[1], args[0]])

    def _deallocate(self):
        """Remove the temp directory from the filesystem recursively"""
        self.remove(recursive=True, force=True)


class Fs:
    """filesystem utils"""

    def __init__(self, client: "CoreClient"):
        self._client = client

    def _cp_dir(self, source: Path, dest: Path, force: bool):
        if not dest.exists():
            dest.mkdir(source.lstat().st_mode)

        files = source.iterdir()

        for src_file in files:
            dest_file = dest / src_file.name

            src_lstat = src_file.lstat()
            if stat.S_ISDIR(src_lstat.st_mode):
                self._cp_dir(src_file, dest_file, force)
            elif stat.S_ISLNK(src_lstat.st_mode):
                symlink_full = src_file.readlink()
                dest_file.symlink_to(symlink_full)
            elif dest_file.exists() and not force:
                pass
            else:
                dest_file.write_bytes(src_file.read_bytes())

    def cp(self, sources: list[Path], dest: Path, recursive: bool, force: bool):
        dest_exists = dest.exists()
        is_dest_dir = dest_exists and dest.is_dir()

        if (not dest_exists or not is_dest_dir) and (len(sources) > 1):
            raise ArgumentError(f"target {dest} is not a directory")

        if recursive and not dest_exists:
            try:
                dest.mkdir(0o777)
            except Exception:
                if not dest.exists():
                    raise

        for source in sources:
            if not source.exists():
                raise ArgumentError(f"cannot stat {source}: No such file or directory")

            if source.is_dir():
                if not recursive:
                    logger.info(f"omitting directory {source}")
                else:
                    cur_dest = dest / source.name
                    source_mode = source.stat().st_mode
                    if not cur_dest.exists():
                        cur_dest.mkdir(source_mode)
                    self._cp_dir(source, cur_dest, force)
            else:  # source is a file
                cur_dest = dest
                if dest.exists() and dest.is_dir():
                    cur_dest = dest / source.name
                if not cur_dest.exists() or force:
                    cur_dest.write_bytes(source.read_bytes())

    @path_to_str("path")
    def is_file(self, path: str) -> bool:
        """Return True if the entry is a file"""
        return bool(self.stat(path).st_mode & S_IFREG)

    @path_to_str("path")
    def _chown(self, path: str, uid: int, gid: int):
        """Change owner and group for a remote path."""
        if self._client.symbols.chown(path, uid, gid).c_int32 < 0:
            self._client.raise_errno_exception(f"failed to chown: {path}")

    @path_to_str("path")
    def chown(self, path: str, uid: int, gid: int, recursive=False):
        """Change owner and group for a path, optionally recursively."""
        if not recursive:
            self._chown(path, uid, gid)
            return

        for file in self.find(path, topdown=False):
            self._chown(file, uid, gid)

    @path_to_str("path")
    def _chmod(self, path: str, mode: int):
        """Change mode bits for a remote path."""
        if self._client.symbols.chmod(path, mode).c_int32 < 0:
            self._client.raise_errno_exception(f"failed to chmod: {path}")

    @path_to_str("path")
    def chmod(self, path: str, mode: int, recursive=False):
        """Change mode bits for a path, optionally recursively."""
        if not recursive:
            self._chmod(path, mode)
            return

        for file in self.find(path, topdown=False):
            self._chmod(file, mode)

    @path_to_str("path")
    def _remove(self, path: str, force=False):
        """Remove a file on the remote filesystem."""
        if self._client.symbols.remove(path).c_int32 < 0 and not force:
            self._client.raise_errno_exception(f"failed to remove: {path}")

    @path_to_str("path")
    def remove(self, path: str, recursive=False, force=False):
        """Remove a file or directory tree on the remote filesystem."""
        if not recursive or self.is_file(path):
            self._remove(path, force=force)
            return

        for filename in self.find(path, topdown=False):
            self._remove(filename, force=force)

    @path_to_str("old")
    @path_to_str("new")
    def rename(self, old: str, new: str):
        """Rename or move a path on the remote filesystem."""
        if self._client.symbols.rename(old, new).c_int32 < 0:
            self._client.raise_errno_exception(f"failed to rename: {old} -> {new}")

    @path_to_str("path")
    def _mkdir(self, path: str, mode: int = 0o777):
        """Create a directory on the remote filesystem."""
        if self._client.symbols.mkdir(path, mode).c_int64 < 0:
            self._client.raise_errno_exception(f"failed to mkdir: {path}")

        # os may not always respect the permission given by the mode argument to mkdir
        self.chmod(path, mode)

    @path_to_str("path")
    def mkdir(self, path: str, mode: int = 0o777, parents=False, exist_ok=False):
        """Create a directory, optionally creating parent directories."""
        if not parents:
            try:
                self._mkdir(path, mode=mode)
            except RpcFileExistsError:
                if not exist_ok:
                    raise
            return

        dir_path = Path(self.pwd())
        for part in Path(path).parts:
            dir_path = dir_path / part
            with contextlib.suppress(RpcIsADirectoryError, RpcFileExistsError):
                self._mkdir(dir_path, mode=mode)

    @path_to_str("path")
    def chdir(self, path: str):
        """Change the remote process working directory."""
        if self._client.symbols.chdir(path).c_int64 < 0:
            self._client.raise_errno_exception(f"failed to chdir: {path}")

    @path_to_str("path")
    def readlink(self, path: str, absolute=True) -> str:
        """Read the symlink target on the remote filesystem."""
        with self._client.safe_calloc(MAXPATHLEN) as buf:
            if self._client.symbols.readlink(path, buf, MAXPATHLEN).c_int64 < 0:
                self._client.raise_errno_exception(f"readlink failed for: {path}")
            if absolute:
                return str(Path(path).parent / buf.peek_str())
            return buf.peek_str()

    @path_to_str("path")
    def realpath(self, path: str) -> str:
        """Resolve a path on the remote filesystem to an absolute path."""
        with self._client.safe_malloc(MAXPATHLEN) as buf:
            if self._client.symbols.realpath(path, buf) == 0:
                self._client.raise_errno_exception(f"realpath failed for: {path}")
            return buf.peek_str()

    @path_to_str("path")
    def is_symlink(self, path: str) -> bool:
        try:
            self.readlink(path)
        except BadReturnValueError:
            return False
        return True

    @path_to_str("file")
    def open(self, file: str, mode: str, access: int = 0o777) -> File:
        """
        call open(file, mode, access) at remote and get a context manager
        file object
        :param file: filename to be opened
        :param mode: one of:
            'r' - read only
            'r+' - read and write. exception if file doesn't exist
            'rw' - read and write. create if it doesn't exist. also truncate.
            'w' - write only. create if it doesn't exist. also truncate.
            'w+' - read and write. create if it doesn't exist.
        :param access: access mode as octal value
        :return: a context manager file object
        """
        available_modes = {
            "r": O_RDONLY,
            "r+": O_RDWR,
            "rw": O_RDWR | O_CREAT | O_TRUNC,
            "w": O_WRONLY | O_CREAT | O_TRUNC,
            "w+": O_RDWR | O_CREAT,
        }
        mode_int = available_modes.get(mode)
        if mode_int is None:
            raise ArgumentError(f"mode can be only one of: {available_modes.keys()}")

        fd = self._client.symbols.open(file, mode_int, access, va_list_index=2).c_int32
        if fd < 0:
            self._client.raise_errno_exception(f"failed to open: {file}")

        return File(self._client, fd)

    @path_to_str("file")
    def write_file(self, file: str, buf: bytes, access: int = 0o777):
        with self.open(file, "w", access=access) as f:
            f.write(buf)

    @path_to_str("file")
    def read_file(self, file: str) -> bytes:
        with self.open(file, "r") as f:
            return f.read()

    def remote_path(self, path: str) -> RemotePath:
        return RemotePath(path, self._client)

    @path_to_str("local")
    def pull(
        self,
        remotes: Union[list[Union[str, Path]], Union[str, Path]],
        local: str,
        recursive: bool = False,
        force: bool = False,
    ):
        """pull complete directory tree"""
        if not isinstance(remotes, list):
            remotes = [posixpath.expanduser(remotes)]
        remotes_str = [posixpath.expanduser(remote) for remote in remotes]
        self.cp([self.remote_path(remote) for remote in remotes_str], Path(str(local)), recursive, force)

    @path_to_str("remote")
    def push(
        self,
        local_files: Union[list[Union[str, Path]], Union[str, Path]],
        remote: str,
        recursive: bool = False,
        force: bool = False,
    ):
        """push complete directory tree"""
        if not isinstance(local_files, list):
            local_files = [posixpath.expanduser(local_files)]
        locals_str = [posixpath.expanduser(local) for local in local_files]
        self.cp([Path(str(local)) for local in locals_str], self.remote_path(remote), recursive, force)

    @path_to_str("file")
    def touch(self, file: str, mode: int = 0o666, exist_ok=True):
        """simulate unix touch command for given file"""
        if not exist_ok:
            try:
                self.stat(file)
                raise RpcFileExistsError()
            except RpcFileNotFoundError:
                pass
        with self.open(file, "w+", mode):
            pass

    @path_to_str("src", "dst")
    def symlink(self, src: str, dst: str) -> int:
        """Create a symbolic link on the remote filesystem."""
        err = self._client.symbols.symlink(src, dst).c_int64
        if err < 0:
            self._client.raise_errno_exception(f"symlink failed to create link: {dst}->{src}")
        return err

    @path_to_str("src", "dst")
    def link(self, src: str, dst: str) -> int:
        """Create a hard link on the remote filesystem."""
        err = self._client.symbols.link(src, dst).c_int64
        if err < 0:
            self._client.raise_errno_exception(f"link failed to create link: {dst}->{src}")
        return err

    def pwd(self) -> str:
        """calls getcwd(buf, size_t) and prints current path.
        with the special values NULL, 0 the buffer is allocated dynamically"""
        chunk = self._client.symbols.getcwd(0, 0)
        if chunk == 0:
            self._client.raise_errno_exception("getcwd() failed")
        buf = chunk.peek_str()
        self._client.symbols.free(chunk)
        return buf

    @path_to_str("path")
    def listdir(self, path: str = ".") -> list[str]:
        """get directory listing for a given dirname"""
        return [e.name for e in self.scandir(path)]

    @path_to_str("path")
    def scandir(self, path: str = ".") -> list[DirEntry]:
        """get directory listing for a given dirname"""
        result = []
        for entry in self._client.listdir(path)[2:]:
            result.append(DirEntry(path, entry, self._client))
        return result

    @path_to_str("path")
    def stat(self, path: str):
        """Return stat info for a remote path (platform-specific implementation)."""
        raise NotImplementedError()

    @path_to_str("path")
    def lstat(self, path: str):
        """Return lstat info for a remote path (platform-specific implementation)."""
        raise NotImplementedError()

    @path_to_str("path")
    def accessible(self, path: str, mode: int = R_OK):
        """check if a given path can be accessed."""
        err = self._client.symbols.access(path, mode)
        return err == 0

    @path_to_str("path")
    def chflags(self, path: str, flags: int = 0):
        """set file flags"""
        err = self._client.symbols.chflags(path, flags)
        if err < 0:
            self._client.raise_errno_exception(f"failed to chflags on: {path}")

    @path_to_str("top")
    def find(self, top: str, topdown=True):
        """traverse a file tree top to down"""
        if topdown:
            if self.accessible(top):
                yield top
            else:
                raise RpcFileNotFoundError(f"cannot access: {top}")

        for root, dirs, files in self.walk(top, topdown=topdown):
            for name in files:
                yield os.path.join(root, name)
            for name in dirs:
                yield os.path.join(root, name)

        if not topdown:
            if self.accessible(top):
                yield top
            else:
                raise RpcFileNotFoundError(f"cannot access: {top}")

    @path_to_str("top")
    def walk(self, top: str, topdown=True, onerror=None):
        """provides the same results as os.walk(top)"""
        dirs = []
        files = []
        try:
            for entry in self.scandir(top):
                try:
                    if entry.is_dir():
                        dirs.append(entry.name)
                    else:
                        files.append(entry.name)
                except Exception as e:
                    if not onerror:
                        raise
                    onerror(e)
        except Exception as e:
            if not onerror:
                raise
            onerror(e)

        if topdown:
            yield top, dirs, files

        if dirs:
            for d in dirs:
                yield from self.walk(posixpath.join(top, d), topdown=topdown, onerror=onerror)

        if not topdown:
            yield top, dirs, files

    @contextlib.contextmanager
    @path_to_str("remote")
    def remote_file(self, remote: str):
        with tempfile.TemporaryDirectory() as local_dir:
            local = Path(local_dir) / Path(remote).parts[-1]
            if self.accessible(remote):
                self.pull(remote, local.absolute())
            try:
                yield local.absolute()
            finally:
                self.push(local, remote, force=True)

    @path_to_str("directory")
    def remote_temp_dir(self, directory: str = "/tmp", mode: int = 0o700) -> RemoteTemporaryDir:
        """Generate a random temp directory name and create it in the given directory (default '/tmp').

        :param directory: Directory to create the temp directory inside (default '/tmp').
        :param mode: Mode to create the temp directory with (default 0o700).
        """
        return RemoteTemporaryDir(self._client, directory, mode)



from abc import ABC, abstractmethod
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Optional

from rpcclient.clients.darwin.common import CfSerializable
from rpcclient.clients.darwin.symbol import DarwinSymbol
from rpcclient.core.allocated import Allocated
from rpcclient.exceptions import RpcClientException

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient


class Duet:
    def __init__(self, client: "DarwinClient") -> None:
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client
        self._client.load_framework("DuetActivityScheduler")
        self.NSArray_cls = client.symbols.objc_getClass("NSArray")
        self.CDPaths_cls = client.symbols.objc_getClass("_CDPaths")
        self.DKEventQuery_cls = client.symbols.objc_getClass("_DKEventQuery")
        self.DKKnowledgeStore_cls = client.symbols.objc_getClass("_DKKnowledgeStore")
        self.DKKnowledgeStorage_cls = client.symbols.objc_getClass("_DKKnowledgeStorage")
        self.DKSystemEventStreams_cls = client.symbols.objc_getClass("_DKSystemEventStreams")
        self.DKSystemEventStreams_cls_cls = client.symbols.object_getClass(self.DKSystemEventStreams_cls)

    def knowledge_store(self) -> "KnowledgeStoreDup":
        """Copy the on device knowledge store to a temp directory and open it read only."""
        return KnowledgeStoreDup(self._client, self)

    def knowledge_store_xpc(self) -> "KnowledgeStoreXPC":
        """Connect directly to the live knowledge store via XPC."""
        return KnowledgeStoreXPC(self._client, self)


class DKEvent:
    """Thin, *pythonic* wrapper around a private `_DKEvent` ObjectiveC object."""

    def __init__(self, event: DarwinSymbol) -> None:
        """Initialise the wrapper.

        :param event: Underlying `_DKEvent` ObjectiveC event.
        """
        self.native = event

    @cached_property
    def start(self) -> datetime:
        """Return the event's *start* timestamp.

        :returns: Event start time.
        """
        return self.native.objc_call("startDate").py()

    @cached_property
    def end(self) -> datetime:
        """Return the event's *end* timestamp (may equal *start*).

        :returns: Event end time.
        """
        return self.native.objc_call("endDate").py()

    @cached_property
    def metadata(self) -> str:
        """Return the event's metadata dictionary, if present.

        :returns: Metadata mapping or `None` when absent.
        """
        return self.native.objc_call("metadata").cfdesc

    @cached_property
    def stream(self) -> str:
        """Return the name of the stream this event belongs to.

        :returns: Stream name.
        """
        return self.native.objc_call("stream").objc_call("name").py()

    @cached_property
    def value(self) -> CfSerializable:
        """Return the primary value associated with the event.

        :returns: Stream specific value type.
        """
        return self.native.objc_call("value").objc_call("primaryValue").py()

    def __str__(self) -> str:
        """Pretty print the event for quick inspection.

        :returns: Formatted multi line string representation.
        """
        output = f"DKEvent - {self.stream}\n\n"
        output += f" value:\t{self.value}\n"
        output += f" start:\t{self.start}\n"
        output += f" end:\t{self.end}"
        output += f" metadata:\n{self.metadata}"
        return output


class KnowledgeStoreContext(ABC, Allocated):
    """Abstract context manager that exposes helper queries for `_DKKnowledgeStore`."""

    def __init__(self, client: "DarwinClient", duet: Duet) -> None:
        """Collect available system event streams and store the *rpcclient* handle.

        :param rpcclient.darwin.client.DarwinClient client:
        :param duet:
        """
        Allocated.__init__(self)
        self._client = client
        self._duet = duet
        self.streams: dict[str, DarwinSymbol] = {}
        with client.safe_malloc(8) as count:
            methods = client.symbols.class_copyMethodList(duet.DKSystemEventStreams_cls_cls, count)
            for i in range(count[0]):
                method_name = client.symbols.sel_getName(client.symbols.method_getName(methods[i])).peek_str()
                stream = duet.DKSystemEventStreams_cls.objc_call(method_name)
                self.streams[stream.objc_call("name").py()] = stream

    def query_events(self, stream: str) -> list["DKEvent"]:
        """Retrieve all events for a single stream.

        :param stream: Stream identifier.
        :returns: List of events or `None` when the stream is empty.
        """
        raw_events = self._query_raw_events(stream)
        event_count = raw_events.objc_call("count") if raw_events != 0 else 0
        return (
            [DKEvent(raw_events.objc_call("objectAtIndex:", i)) for i in range(event_count)] if event_count != 0 else []
        )

    def query_events_streams(self, streams: Optional[list[str]] = None) -> dict[str, list["DKEvent"]]:
        """Query multiple streams at once.

        :param streams: Specific streams to query. `None` queries all known streams.
        :returns: Mapping of stream name to event list (absent when stream empty).
        """
        events: dict[str, list[DKEvent]] = {}
        streams = list(self.streams.keys()) if streams is None else streams
        for stream in streams:
            events_for_stream = self.query_events(stream)
            if len(events_for_stream) != 0:
                events[stream] = events_for_stream
        return events

    def _query_raw_events(self, stream: str) -> DarwinSymbol:
        """Build and execute a `_DKEventQuery` limited to *stream*.

        :param stream: Stream name.
        :returns: ObjectiveC array containing raw `_DKEvent` objects.
        """
        query = self._duet.DKEventQuery_cls.objc_call("new")
        try:
            with self._client.safe_malloc(8) as buf:
                buf[0] = self.streams[stream]
                arr = self._duet.NSArray_cls.objc_call("arrayWithObjects:count:", buf, 1)
                query.objc_call("setEventStreams:", arr)
                return self._execute_query(query)
        finally:
            query.objc_call("release")

    def _execute_query(self, query: DarwinSymbol) -> DarwinSymbol:
        """Run *query* through the concrete :pyattr:`knowledge_store` implementation.

        :param query: Configured `_DKEventQuery` instance.
        :returns: ObjectiveC array with results.
        :raises RpcClientException: If :pyattr:`knowledge_store` is not yet initialized.
        """
        if self.knowledge_store is None:
            raise RpcClientException("KnowledgeStore not initialized")
        with self._client.safe_malloc(8) as error:
            return self.knowledge_store.objc_call("executeQuery:error:", query, error)

    def _deallocate(self):
        """Release the allocated knowledge store object"""
        self.knowledge_store.objc_call("release")

    @property
    @abstractmethod
    def knowledge_store(self) -> DarwinSymbol:
        """Return the clients specific `_DKKnowledgeStore` handle.

        :returns: Concrete store.
        """
        pass


class KnowledgeStoreDup(KnowledgeStoreContext):
    """Copy the on device knowledge store to a temp directory and open it read only."""

    def __init__(self, client: "DarwinClient", duet: Duet) -> None:
        """Create the temp directory, copy knowledge store files inside and init knowledge store objects.

        :param rpcclient.darwin.client.DarwinClient client:
        :param duet:
        """
        super().__init__(client, duet)
        origin_path = client.fs.remote_path(duet.CDPaths_cls.objc_call("knowledgeDirectory").py())
        self.tmp_dir = client.fs.remote_temp_dir()
        tmp_knowledge_dir = str(self.tmp_dir / origin_path.name)
        try:
            client.fs.cp([origin_path], self.tmp_dir, recursive=True, force=False)
            knowledge_storage = self._duet.DKKnowledgeStorage_cls.objc_call(
                "storageWithDirectory:readOnly:", self._client.cf(tmp_knowledge_dir), 1
            )
            knowledge_store = self._duet.DKKnowledgeStore_cls.objc_call("alloc")
            self._knowledge_store = knowledge_store.objc_call(
                "initWithKnowledgeStoreHandle:readOnly:", knowledge_storage, 1
            )
        except Exception:
            self.tmp_dir.deallocate()
            raise

    def _deallocate(self):
        """Delete the allocated temp dir and release the allocated knowledge store object"""
        self.tmp_dir.deallocate()
        super()._deallocate()

    @property
    def knowledge_store(self) -> Optional[DarwinSymbol]:
        """Expose the read only copied knowledge store handle.

        :returns: Store handle or `None` if context not entered.
        """
        return self._knowledge_store


class KnowledgeStoreXPC(KnowledgeStoreContext):
    """Connect directly to the live knowledge store via XPC."""

    def __init__(self, client: "DarwinClient", duet: Duet) -> None:
        """connect to the XPC knowledge store interface on the device.

        :param rpcclient.darwin.client.DarwinClient client:
        :param duet:
        """
        super().__init__(client, duet)
        self._knowledge_store: DarwinSymbol = duet.DKKnowledgeStore_cls.objc_call("knowledgeStore")

    @property
    def knowledge_store(self) -> Optional[DarwinSymbol]:
        """Expose the live XPC backed knowledge store handle.

        :returns: Store handle.
        """
        return self._knowledge_store



import ast
import builtins
from typing import Any

from IPython.terminal.interactiveshell import TerminalInteractiveShell

from rpcclient.clients.darwin.objective_c import objective_c_class
from rpcclient.core.symbols_jar import SymbolsJar
from rpcclient.exceptions import GettingObjectiveCClassError, SymbolAbsentError


class RpcEvents:
    """Small IPython event handler that lazy-loads symbols and Obj-C classes."""

    def __init__(self, ip: TerminalInteractiveShell) -> None:
        """Store the IPython shell reference."""
        self.ipython = ip

    def pre_run_cell(self, info: Any) -> None:
        """
        Before each cell: inject missing names by lazy-loading SymbolsJar or Objective-C classes.
        Expects `info.raw_cell` to contain the cell source.
        """
        client = self.ipython.user_ns.get("p")
        if client is None:
            return

        # Skip parsing of this cell if it's an IPython's magic transformation (like '!', '?', '%')
        if info.raw_cell != info.transformed_cell.rstrip("\n"):
            return

        for node in ast.walk(ast.parse(info.raw_cell)):
            if not isinstance(node, ast.Name):
                continue

            existing = self.ipython.user_ns.get(node.id)
            objc_class_place_holder = (
                isinstance(existing, objective_c_class.Class) and getattr(existing, "name", "") == ""
            )

            # Skip names already known to local/global/builtins.
            if (
                node.id in locals()
                or node.id in globals()
                or node.id in dir(builtins)
                or (existing is not None and not objc_class_place_holder)
            ):
                continue

            # 1) Generic: SymbolsJar lazy load
            if not hasattr(SymbolsJar, node.id):
                try:
                    symbol = getattr(client.symbols, node.id)
                except SymbolAbsentError:
                    pass
                else:
                    self.ipython.user_ns[node.id] = symbol
                    continue

            # 2) Darwin-specific: Objective-C class autoload / lazy reload
            if hasattr(client, "objc_get_class"):
                if objc_class_place_holder:
                    existing.reload()
                    continue

                try:
                    objc_cls = client.objc_get_class(node.id)
                except GettingObjectiveCClassError:
                    pass
                else:
                    self.ipython.user_ns[node.id] = objc_cls
                    continue


def load_ipython_extension(ip: TerminalInteractiveShell) -> None:
    """IPython extension hook: register the pre_run_cell handler."""
    handlers = RpcEvents(ip)
    ip.events.register("pre_run_cell", handlers.pre_run_cell)



def test_duet_knowledge_store(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.duet.knowledge_store() as ctx:
        events = ctx.query_events_streams()
        assert events.keys() != 0



import datetime
import logging
import re
from functools import cached_property
from typing import TYPE_CHECKING, Optional

from rpcclient.clients.darwin.consts import OsLogLevel
from rpcclient.clients.darwin.subsystems.cfpreferences import GLOBAL_DOMAIN, kCFPreferencesAnyHost
from rpcclient.clients.darwin.subsystems.processes import Process
from rpcclient.clients.darwin.symbol import DarwinSymbol
from rpcclient.core.structs.consts import SIGKILL
from rpcclient.exceptions import BadReturnValueError, HarGlobalNotFoundError, MissingLibraryError

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient

MOV_RDI_RBX = b"\x48\x89\xdf"
XOR_ESI_ESI = b"\x31\xf6"
MOV_EDX_0A = b"\xba\x0a\x00\x00"
INTEL_PATTERN = MOV_RDI_RBX + XOR_ESI_ESI + MOV_EDX_0A + b"\x00"

MOV_X0_X9 = b"\xe0\x03\x13\xaa"
MOV_X1_0 = b"\x01\x00\x80\xd2"
MOV_W2_0XA = b"\x42\x01\x80\x52"
ARM_PATTERN = MOV_X0_X9 + MOV_X1_0 + MOV_W2_0XA

logger = logging.getLogger(__name__)


class OsLogPreferencesBase:
    def __init__(self, client: "DarwinClient", obj: DarwinSymbol):
        self._client = client
        self._object = obj

    @property
    def name(self) -> str:
        return self._object.objc_call("name").py()

    @property
    def persisted_level(self) -> OsLogLevel:
        return OsLogLevel(self._object.objc_call("persistedLevel"))

    @persisted_level.setter
    def persisted_level(self, value: OsLogLevel) -> None:
        self._object.objc_call("setPersistedLevel:", value)

    @property
    def default_persisted_level(self) -> OsLogLevel:
        return OsLogLevel(self._object.objc_call("defaultPersistedLevel"))

    @property
    def enabled_level(self) -> OsLogLevel:
        return OsLogLevel(self._object.objc_call("enabledLevel"))

    @enabled_level.setter
    def enabled_level(self, value: OsLogLevel) -> None:
        self._object.objc_call("setEnabledLevel:", value)

    @property
    def default_enabled_level(self) -> OsLogLevel:
        return OsLogLevel(self._object.objc_call("defaultEnabledLevel"))

    @property
    def _verbosity_description(self) -> str:
        return (
            f"ENABLED_LEVEL:{self.enabled_level.name}({self.enabled_level.value}) "
            f"PERSISTED_LEVEL:{self.persisted_level.name}({self.persisted_level.value})"
        )

    def reset(self) -> None:
        self._object.objc_call("reset")

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} NAME:{self.name} {self._verbosity_description}>"


class OsLogPreferencesCategory(OsLogPreferencesBase):
    def __init__(self, client: "DarwinClient", category: str, subsystem: DarwinSymbol):
        obj = (
            client.symbols.objc_getClass("OSLogPreferencesCategory")
            .objc_call("alloc")
            .objc_call("initWithName:subsystem:", client.cf(category), subsystem)
        )
        super().__init__(client, obj)

    @property
    def name(self) -> str:
        return self._object.objc_call("name").py()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} NAME:{self.name} {self._verbosity_description}>"


class OsLogPreferencesSubsystem(OsLogPreferencesBase):
    def __init__(self, client: "DarwinClient", subsystem: str):
        obj = (
            client.symbols.objc_getClass("OSLogPreferencesSubsystem")
            .objc_call("alloc")
            .objc_call("initWithName:", client.cf(subsystem))
        )
        super().__init__(client, obj)

    @property
    def category_strings(self) -> list[OsLogPreferencesCategory]:
        return self._object.objc_call("categories").py()

    @property
    def categories(self) -> list[OsLogPreferencesCategory]:
        result = []
        for name in self.category_strings:
            result.append(self.get_category(name))
        return result

    def get_category(self, name: str) -> OsLogPreferencesCategory:
        return OsLogPreferencesCategory(self._client, name, self._object)


class OsLogPreferencesManager(OsLogPreferencesBase):
    def __init__(self, client: "DarwinClient"):
        obj = client.symbols.objc_getClass("OSLogPreferencesManager").objc_call("sharedManager")
        super().__init__(client, obj)

    @property
    def subsystem_strings(self) -> list[str]:
        return self._object.objc_call("subsystems").py()

    @property
    def subsystems(self) -> list[OsLogPreferencesSubsystem]:
        result = []
        for name in self.subsystem_strings:
            result.append(self.get_subsystem(name))
        return result

    def get_subsystem(self, name: str) -> OsLogPreferencesSubsystem:
        return OsLogPreferencesSubsystem(self._client, name)


class Syslog:
    """ " manage syslog"""

    def __init__(self, client: "DarwinClient"):
        """
        @type client: rpcclient.darwin.client.DarwinClient
        """
        self._client = client
        self._client.load_framework("LoggingSupport")
        self.preferences_manager = OsLogPreferencesManager(self._client)

    @cached_property
    def _cfnetwork_base(self) -> int:
        self._client.load_framework("CFNetwork")
        cfnetwork = [
            image
            for image in self._client.images
            if image.name.startswith("/System/Library/Frameworks/CFNetwork.framework")
        ]
        if len(cfnetwork) < 1:
            raise MissingLibraryError()
        return cfnetwork[0].base_address

    @cached_property
    def _arm_enable_har_global(self) -> DarwinSymbol:
        """
        In order to find an unexported global variable, we use a sequence of instructions that we know comes after
        the aforementioned variable is accessed.

            88 07 2D D0                 ADRP            X8, #har_global@PAGE
            08 41 55 39                 LDRB            W8, [X8,#har_global@PAGEOFF]
            08 03 00 34                 CBZ             W8, loc_187DA1294
            E0 03 13 AA                 MOV             X0, X19 ; __str
            01 00 80 D2                 MOV             X1, #0  ; __endptr
            42 01 80 52                 MOV             W2, #0xA ; __base

        After getting the address of the instruction `mov x0,x19`, we need to rewind 2 instructions. Next, we verify
        that the above 2 instructions are `adrp` && `ldrb` . Then, extract the page address + page offset and
        calculate the affective address.
        """

        address = []
        regex_hex = r"0x[0-9a-fA-F]+"
        start = self._cfnetwork_base
        while True:
            pattern_addr = self._client.symbols.memmem(start, 0xFFFFFFFF, ARM_PATTERN, len(ARM_PATTERN))
            if pattern_addr == 0:
                raise HarGlobalNotFoundError()
            pattern_sym = self._client.symbol(pattern_addr)
            disass = (pattern_sym - len(ARM_PATTERN)).disass(8)
            if disass[0].mnemonic == "adrp" and disass[1].mnemonic == "ldrb":
                break
            # search next occurrence
            start = pattern_addr + 1

        for instruction in disass:
            address.append(int(re.findall(regex_hex, instruction.op_str)[0], 16))
        return self._client.symbol(sum(address))

    @cached_property
    def _intel_enable_har_global(self) -> DarwinSymbol:
        """
        In order to find an unexported global variable, we use a sequence of instructions that we know comes before
        the aforementioned variable is accessed.

            48 89 DF                    mov     rdi, rbx        ; __str
            31 F6                       xor     esi, esi        ; __endptr
            BA 0A 00 00                 mov     edx, 0Ah        ; __base
            00
            E8 78 EE 10                 call    _strtol
            00
            48 89 C3                    mov     rbx, rax
            88 1D 19 9C                 mov     cs:har_global_0, bl  <-----------------
        """
        op_str_pattern = "byte ptr [rip +"
        start = self._cfnetwork_base
        while True:
            pattern_addr = self._client.symbols.memmem(start, 0xFFFFFFFF, INTEL_PATTERN, len(INTEL_PATTERN))
            if pattern_addr == 0:
                raise HarGlobalNotFoundError()
            pattern_sym = self._client.symbol(pattern_addr)
            disass = (pattern_sym + len(INTEL_PATTERN)).disass(100)
            if (
                disass[0].mnemonic == "call"
                and disass[1].mnemonic == "mov"
                and disass[2].mnemonic == "mov"
                and disass[2].op_str.startswith(op_str_pattern)
            ):
                offset = int(disass[2].op_str.split(op_str_pattern, 1)[1].split("]")[0], 0)
                return self._client.symbol(disass[2].address + offset)
            start = pattern_addr + 1

    @cached_property
    def _enable_har_global(self) -> DarwinSymbol:
        if self._client.uname.machine == "x86_64":
            return self._intel_enable_har_global
        return self._arm_enable_har_global

    def install_cfnetwork_diagnostics_profile(self) -> None:
        self._client.preferences.cf.set(
            "AppleCFNetworkDiagnosticLogging", 3, GLOBAL_DOMAIN, username="kCFPreferencesAnyUser"
        )

    def remove_cfnetwork_diagnostics_profile(self) -> None:
        self._client.preferences.cf.remove(
            "AppleCFNetworkDiagnosticLogging", GLOBAL_DOMAIN, username="kCFPreferencesAnyUser"
        )

    def set_harlogger_for_process(self, value: bool, pid: int) -> None:
        process = self._client.processes.get_by_pid(pid)
        self._set_harlogger_for_process(value, process)
        self.set_har_capture_global(True)

    def set_harlogger_for_all(self, value: bool, expression: Optional[str] = None) -> None:
        for p in self._client.processes.list():
            if p.pid == 0:
                continue
            if expression and expression not in p.basename:
                continue
            try:
                self._set_harlogger_for_process(value, p)
                logger.info(f"{'Enabled' if value else 'Disabled'} for {p.basename}")
            except BadReturnValueError:
                logger.exception(f"Failed To enable for {p}")
        self.set_har_capture_global(True)

    def set_unredacted_logs(self, enable: bool = True):
        """
        enable/disable unredacted logs (allows seeing the <private> strings)
        https://github.com/EthanArbuckle/unredact-private-os_logs
        """
        with self._client.preferences.sc.open("/Library/Preferences/Logging/com.apple.system.logging.plist") as pref:
            pref.set_dict({"Enable-Logging": True, "Enable-Private-Data": enable})
        self._client.processes.get_by_basename("logd").kill(SIGKILL)

    def set_har_capture_global(self, enable: bool = True):
        """
        enable/disable HAR logging
        https://github.com/doronz88/harlogger
        """
        users = ["mobile", "root"]
        if enable:
            for user in users:
                # settings copied from DiagnosticExtension.appex
                self._client.preferences.cf.set(
                    "har-capture-global",
                    datetime.datetime(9999, 12, 31, 23, 59, 59),
                    "com.apple.CFNetwork",
                    user,
                    hostname=kCFPreferencesAnyHost,
                )
                self._client.preferences.cf.set(
                    "har-body-size-limit", 0x100000, "com.apple.CFNetwork", user, hostname=kCFPreferencesAnyHost
                )
                subsystem = self.preferences_manager.get_subsystem("com.apple.CFNetwork")
                category = subsystem.get_category("HAR")

                category.enabled_level = 2
                category.persisted_level = 2
        else:
            for user in users:
                self._client.preferences.cf.set(
                    "har-capture-global",
                    datetime.datetime(1970, 1, 1, 1, 1, 1),
                    "com.apple.CFNetwork",
                    user,
                    hostname=kCFPreferencesAnyHost,
                )

        if self._client.symbols.notify_post("com.apple.CFNetwork.har-capture-update"):
            raise BadReturnValueError("notify_post() failed")

    def _set_harlogger_for_process(self, value: bool, process: Process) -> None:
        process.poke(self._enable_har_global, value.to_bytes(1, "little"))



import pytest

STDOUT_FD = 1
STDERR_FD = 2
TEST_DATA: bytes = b"testtesttest"


@pytest.mark.parametrize("fd", [STDOUT_FD, STDERR_FD])
def test_capture_fd(client, fd: int) -> None:
    """
    :param rpcclient.client.Client client:
    :param int fd:
    """
    with client.capture_fd(fd) as cap:
        if client.symbols.write(fd, TEST_DATA, len(TEST_DATA)) == -1:
            client.raise_errno_exception("write to fd failed")
        data = cap.read()
    assert data == TEST_DATA


@pytest.mark.parametrize("fd", [STDOUT_FD, STDERR_FD])
def test_fd_cleanup(client, fd: int) -> None:
    """
    :param rpcclient.client.Client client:
    :param int fd:
    """
    fd_count = len(client.processes.get_self().fds)
    with client.capture_fd(fd):
        pass
    assert fd_count == len(client.processes.get_self().fds)



import json
import logging
import plistlib
from collections import namedtuple
from dataclasses import dataclass
from functools import cache, cached_property

from IPython import get_ipython
from tqdm import tqdm

from rpcclient.clients.darwin.common import CfSerializable
from rpcclient.clients.darwin.consts import CFPropertyListFormat, CFPropertyListMutabilityOptions, kCFAllocatorDefault
from rpcclient.clients.darwin.objective_c import autorelease_pool, objective_c_class
from rpcclient.clients.darwin.objective_c.objective_c_symbol import ObjectiveCSymbol
from rpcclient.clients.darwin.structs import utsname
from rpcclient.clients.darwin.subsystems.biome import Biome
from rpcclient.clients.darwin.subsystems.bluetooth import Bluetooth
from rpcclient.clients.darwin.subsystems.core_graphics import CoreGraphics
from rpcclient.clients.darwin.subsystems.darwin_lief import DarwinLief
from rpcclient.clients.darwin.subsystems.duet import Duet
from rpcclient.clients.darwin.subsystems.fs import DarwinFs
from rpcclient.clients.darwin.subsystems.hid import Hid
from rpcclient.clients.darwin.subsystems.ioregistry import IORegistry
from rpcclient.clients.darwin.subsystems.keychain import Keychain
from rpcclient.clients.darwin.subsystems.location import Location
from rpcclient.clients.darwin.subsystems.media import DarwinMedia
from rpcclient.clients.darwin.subsystems.network import DarwinNetwork
from rpcclient.clients.darwin.subsystems.power import Power
from rpcclient.clients.darwin.subsystems.preferences import Preferences
from rpcclient.clients.darwin.subsystems.processes import DarwinProcesses
from rpcclient.clients.darwin.subsystems.syslog import Syslog
from rpcclient.clients.darwin.subsystems.time import Time
from rpcclient.clients.darwin.subsystems.xpc import Xpc
from rpcclient.clients.darwin.symbol import DarwinSymbol
from rpcclient.core.client import CoreClient
from rpcclient.core.structs.consts import RTLD_GLOBAL, RTLD_NOW
from rpcclient.core.subsystems.decorator import subsystem
from rpcclient.core.symbol import Symbol
from rpcclient.exceptions import CfSerializationError, MissingLibraryError
from rpcclient.protocol.rpc_bridge import RpcBridge
from rpcclient.protos.rpc_api_pb2 import MsgId

IsaMagic = namedtuple("IsaMagic", "mask value")
ISA_MAGICS = [
    # ARM64
    IsaMagic(mask=0x000003F000000001, value=0x000001A000000001),
    # X86_64
    IsaMagic(mask=0x001F800000000001, value=0x001D800000000001),
]
# Mask for tagged pointer, from objc-internal.h
OBJC_TAG_MASK = 1 << 63

FRAMEWORKS_PATH = "/System/Library/Frameworks"
PRIVATE_FRAMEWORKS_PATH = "/System/Library/PrivateFrameworks"
LIB_PATH = "/usr/lib"

FRAMEWORKS_BLACKLIST = (
    "PowerlogLiteOperators.framework",
    "PowerlogCore.framework",
    "PowerlogHelperdOperators.framework",
    "PowerlogFullOperators.framework",
    "PowerlogAccounting.framework",
    "JavaVM.framework",
    "ActionKit.framework",
    "DashBoard.framework",
    "CoverSheet.framework",
    "StoreKitMacHelper.framework",
    "ReplayKitMacHelper.framework",
    "PassKitMacHelper.framework",
)

logger = logging.getLogger(__name__)


@dataclass
class DyldImage:
    name: str
    base_address: int


class DarwinClient(CoreClient):
    def __init__(self, bridge: RpcBridge):
        super().__init__(bridge, dlsym_global_handle=RTLD_GLOBAL)

        if self.dlopen("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation", RTLD_NOW) == 0:
            raise MissingLibraryError("failed to load CoreFoundation")

        self.loaded_objc_classes = []
        self._NSPropertyListSerialization = self.symbols.objc_getClass("NSPropertyListSerialization")
        self._CFNullTypeID = self.symbols.CFNullGetTypeID()

    @subsystem
    def biome(self) -> Biome:
        return Biome(self)

    @subsystem
    def duet(self) -> Duet:
        return Duet(self)

    @subsystem
    def fs(self) -> DarwinFs:
        return DarwinFs(self)

    @subsystem
    def preferences(self) -> Preferences:
        return Preferences(self)

    @subsystem
    def processes(self) -> DarwinProcesses:
        return DarwinProcesses(self)

    @subsystem
    def media(self) -> DarwinMedia:
        return DarwinMedia(self)

    @subsystem
    def ioregistry(self) -> IORegistry:
        return IORegistry(self)

    @subsystem
    def xpc(self) -> Xpc:
        return Xpc(self)

    @subsystem
    def syslog(self) -> Syslog:
        return Syslog(self)

    @subsystem
    def time(self) -> Time:
        return Time(self)

    @subsystem
    def hid(self) -> Hid:
        return Hid(self)

    @subsystem
    def lief(self) -> DarwinLief:
        return DarwinLief(self)

    @subsystem
    def bluetooth(self) -> Bluetooth:
        return Bluetooth(self)

    @subsystem
    def core_graphics(self) -> CoreGraphics:
        return CoreGraphics(self)

    @subsystem
    def keychain(self) -> Keychain:
        return Keychain(self)

    @subsystem
    def network(self) -> DarwinNetwork:
        return DarwinNetwork(self)

    @subsystem
    def power(self) -> Power:
        return Power(self)

    @subsystem
    def location(self) -> Location:
        return Location(self)

    @property
    def images(self) -> list[DyldImage]:
        m = []
        for i in range(self.symbols._dyld_image_count()):
            module_name = self.symbols._dyld_get_image_name(i).peek_str()
            base_address = self.symbols._dyld_get_image_header(i)
            m.append(DyldImage(module_name, base_address))
        return m

    @cached_property
    def uname(self):
        with self.safe_calloc(utsname.sizeof()) as uname:
            assert self.symbols.uname(uname) == 0
            return utsname.parse_stream(uname)

    @cached_property
    def is_idevice(self):
        return self.uname.machine.startswith("i")

    @property
    def roots(self) -> list[str]:
        """get a list of all accessible darwin roots when used for lookup of files/preferences/..."""
        return ["/", "/var/root"]

    def showobject(self, object_address: Symbol) -> dict:
        return json.loads(self.rpc_call(MsgId.REQ_SHOW_OBJECT, address=object_address).description)

    def showclass(self, class_address: Symbol) -> dict:
        return json.loads(self.rpc_call(MsgId.REQ_SHOW_CLASS, address=class_address).description)

    def get_class_list(self) -> dict[str, objective_c_class.Class]:
        ret = self.rpc_call(MsgId.REQ_GET_CLASS_LIST)
        result = {}
        for _class in ret.classes:
            result[_class.name] = objective_c_class.Class(self, self.symbol(_class.address), lazy=True)
        return result

    def symbol(self, symbol: int):
        """at a symbol object from a given address"""
        return DarwinSymbol.create(symbol, self)

    def decode_cf(self, symbol: Symbol) -> CfSerializable:
        if self.symbols.CFGetTypeID(symbol) == self._CFNullTypeID:
            return None

        with self.safe_malloc(8) as p_error:
            p_error[0] = 0
            objc_data = self._NSPropertyListSerialization.objc_call(
                "dataWithPropertyList:format:options:error:",
                symbol,
                CFPropertyListFormat.kCFPropertyListBinaryFormat_v1_0,
                0,
                p_error,
            )
            if p_error[0] != 0:
                raise CfSerializationError()
        if objc_data == 0:
            return None
        count = self.symbols.CFDataGetLength(objc_data)
        result = plistlib.loads(self.symbols.CFDataGetBytePtr(objc_data).peek(count))
        objc_data.objc_call("release")
        return result

    def cf(self, o: CfSerializable) -> DarwinSymbol:
        """construct a CFObject from a given python object"""
        if o is None:
            return self.symbols.kCFNull[0]

        plist_bytes = plistlib.dumps(o, fmt=plistlib.FMT_BINARY)
        plist_objc_bytes = self.symbols.CFDataCreate(kCFAllocatorDefault, plist_bytes, len(plist_bytes))
        with self.safe_malloc(8) as p_error:
            p_error[0] = 0
            result = self._NSPropertyListSerialization.objc_call(
                "propertyListWithData:options:format:error:",
                plist_objc_bytes,
                CFPropertyListMutabilityOptions.kCFPropertyListMutableContainersAndLeaves,
                0,
                p_error,
            )
            if p_error[0] != 0:
                raise CfSerializationError()
            return result

    def objc_symbol(self, address) -> ObjectiveCSymbol:
        """
        Get objc symbol wrapper for given address
        :param address:
        :return: ObjectiveC symbol object
        """
        return ObjectiveCSymbol.create(int(address), self)

    @cache
    def objc_get_class(self, name: str):
        """
        Get ObjC class object
        :param name:
        :return:
        """
        return objective_c_class.Class.from_class_name(self, name)

    def is_objc_type(self, symbol: DarwinSymbol) -> bool:
        """
        Test if a given symbol represents an objc object
        :param symbol:
        :return:
        """
        class_info = self.processes.get_self().get_symbol_class_info(symbol)
        if class_info == 0:
            return False
        return class_info.objc_call("typeName").py() == "ObjC"

    def rebind_symbols(self, populate_global_scope=True) -> None:
        ip = get_ipython()
        if ip is None:
            raise RuntimeError("ipython is not running")
        logger.debug("rebinding symbols")
        self.loaded_objc_classes.clear()

        # enumerate all loaded objc classes
        for name, class_ in self.get_class_list().items():
            self.loaded_objc_classes.append(name)
            if populate_global_scope:
                ip.user_ns[name] = class_

    def load_framework(self, name: str) -> None:
        lib = self.dlopen(f"{FRAMEWORKS_PATH}/{name}.framework/{name}", RTLD_NOW)
        if lib == 0:
            lib = self.dlopen(f"{PRIVATE_FRAMEWORKS_PATH}/{name}.framework/{name}", RTLD_NOW)
        if lib == 0:
            raise MissingLibraryError(f"failed to load {name}")

    def load_all_libraries(self, rebind_symbols=True) -> None:
        logger.debug(f"loading frameworks: {FRAMEWORKS_PATH}")
        self._load_frameworks(FRAMEWORKS_PATH)
        logger.debug(f"loading frameworks: {PRIVATE_FRAMEWORKS_PATH}")
        self._load_frameworks(PRIVATE_FRAMEWORKS_PATH)

        logger.debug(f"loading libraries: {LIB_PATH}")
        for filename in tqdm(self.fs.listdir(LIB_PATH)):
            if not filename.endswith(".dylib"):
                continue
            self.dlopen(f"{LIB_PATH}/{filename}", RTLD_NOW)

        if rebind_symbols:
            self.rebind_symbols()

    def _load_frameworks(self, frameworks_path: str):
        for filename in tqdm(self.fs.listdir(frameworks_path)):
            if filename in FRAMEWORKS_BLACKLIST:
                continue
            if "SpringBoard" in filename or "UI" in filename:
                continue
            self.dlopen(f"{frameworks_path}/{filename}/{filename.split('.', 1)[0]}", RTLD_NOW)

    def create_autorelease_pool_ctx(self) -> autorelease_pool.AutorelesePoolCtx:
        """
        Create `AutoreleasePoolCtx` representing an Objective-C `NSAutoreleasePool`.
        Automatically initialize the new pool, can be used with `with` to
        manage a section which would be drained after exiting.

        :return: `AutorelesePoolCtx`
        """
        return autorelease_pool.AutorelesePoolCtx(self)

    def get_autorelease_pools(self) -> list[autorelease_pool.AutoreleasePool]:
        """
        Get all autorelease pools currently in the thread.

        :return: List of `AutoreleasePool` instances found in the dump
        """
        return autorelease_pool.get_autorelease_pools(self)

    def get_current_autorelease_pool(self) -> autorelease_pool.AutoreleasePool:
        """
        Get the most recently created autorelease pool.

        :return: The last `AutoreleasePool` in the list (most recent)
        :raises IndexError: if no pools are found
        """
        return autorelease_pool.get_current_autorelease_pool(self)



import typing
from contextlib import suppress
from typing import TYPE_CHECKING

from rpcclient.exceptions import BadReturnValueError, NoSuchPreferenceError, RpcClientException

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient

kCFPreferencesCurrentUser = "kCFPreferencesCurrentUser"
kCFPreferencesAnyUser = "kCFPreferencesAnyUser"
kCFPreferencesCurrentHost = "kCFPreferencesCurrentHost"
kCFPreferencesAnyHost = "kCFPreferencesAnyHost"
GLOBAL_DOMAIN = "Apple Global Domain"


class CFPreferences:
    """
    API to the CFPreferences* functions - preferences managed by cfprefsd.
    https://developer.apple.com/documentation/corefoundation/preferences_utilities?language=objc
    """

    def __init__(self, client: "DarwinClient"):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client

    def get_keys(
        self, application_id: str, username: str = kCFPreferencesCurrentUser, hostname: str = kCFPreferencesCurrentHost
    ) -> typing.Optional[list[str]]:
        """wrapper for CFPreferencesCopyKeyList"""
        application_id = self._client.cf(application_id)
        username = self._client.cf(username)
        hostname = self._client.cf(hostname)
        keys = self._client.symbols.CFPreferencesCopyKeyList(application_id, username, hostname).py()
        if keys is None:
            raise NoSuchPreferenceError()
        return keys

    def get_value(
        self,
        key: str,
        application_id: str,
        username: str = kCFPreferencesCurrentUser,
        hostname: str = kCFPreferencesCurrentHost,
    ) -> typing.Optional[str]:
        """wrapper for CFPreferencesCopyValue"""
        key = self._client.cf(key)
        application_id = self._client.cf(application_id)
        username = self._client.cf(username)
        hostname = self._client.cf(hostname)
        return self._client.symbols.CFPreferencesCopyValue(key, application_id, username, hostname).py()

    def get_dict(
        self, application_id: str, username: str = kCFPreferencesCurrentUser, hostname: str = kCFPreferencesCurrentHost
    ) -> typing.Optional[dict]:
        """get a dictionary representation of given preference"""
        result = {}
        key_list = self.get_keys(application_id, username, hostname)
        if not key_list:
            raise RpcClientException(f"failed to get key list for: {application_id}/{username}/{hostname}")
        for k in key_list:
            result[k] = self.get_value(k, application_id, username, hostname)
        return result

    def set(
        self,
        key: str,
        value,
        application_id: str,
        username: str = kCFPreferencesCurrentUser,
        hostname: str = kCFPreferencesCurrentHost,
    ):
        """wrapper for CFPreferencesSetValue"""
        self._client.symbols.CFPreferencesSetValue(
            self._client.cf(key),
            self._client.cf(value),
            self._client.cf(application_id),
            self._client.cf(username),
            self._client.cf(hostname),
        )

    def remove(
        self,
        key: str,
        application_id: str,
        username: str = kCFPreferencesCurrentUser,
        hostname: str = kCFPreferencesCurrentHost,
    ):
        """remove a given key from a preference"""
        self._client.symbols.CFPreferencesSetValue(
            self._client.cf(key),
            0,
            self._client.cf(application_id),
            self._client.cf(username),
            self._client.cf(hostname),
        )

    def set_dict(
        self,
        d: dict,
        application_id: str,
        username: str = kCFPreferencesCurrentUser,
        hostname: str = kCFPreferencesCurrentHost,
    ):
        """set entire preference dictionary (erase first if exists)"""
        with suppress(NoSuchPreferenceError):
            self.clear(application_id, username, hostname)
        self.update_dict(d, application_id, username, hostname)

    def update_dict(
        self,
        d: dict,
        application_id: str,
        username: str = kCFPreferencesCurrentUser,
        hostname: str = kCFPreferencesCurrentHost,
    ):
        """update preference dictionary"""
        for k, v in d.items():
            self.set(k, v, application_id, username, hostname)

    def clear(
        self, application_id: str, username: str = kCFPreferencesCurrentUser, hostname: str = kCFPreferencesCurrentHost
    ):
        """remove all values from given preference"""
        for k in self.get_keys(application_id, username, hostname):
            self.remove(k, application_id, username, hostname)

    def sync(
        self, application_id: str, username: str = kCFPreferencesCurrentUser, hostname: str = kCFPreferencesCurrentHost
    ):
        if not self._client.symbols.CFPreferencesSynchronize(
            self._client.cf(application_id), self._client.cf(username), self._client.cf(hostname)
        ):
            raise BadReturnValueError("CFPreferencesSynchronize() failed")
